# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
#              C E D A R
#          S O L U T I O N S       "Software done right."
#           S O F T W A R E
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Copyright (c) 2004-2008,2010,2015 Kenneth J. Pronovici.
# All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License,
# Version 2, as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# Copies of the GNU General Public License are available from
# the Free Software Foundation website, http://www.gnu.org/.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Author   : Kenneth J. Pronovici <pronovic@ieee.org>
# Language : Python 3
# Project  : Cedar Backup, release 3
# Purpose  : Provides backup peer-related objects.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

########################################################################
# Module documentation
########################################################################

"""
Provides backup peer-related objects and utility functions.

Module Attributes
=================

Attributes:
   DEF_COLLECT_INDICATOR: Name of the default collect indicator file
   DEF_STAGE_INDICATOR: Name of the default stage indicator file

:author: Kenneth J. Pronovici <pronovic@ieee.org>
"""


########################################################################
# Imported modules
########################################################################

import logging
import os
import posixpath
import shutil
import sys

from CedarBackup3.config import VALID_FAILURE_MODES
from CedarBackup3.filesystem import FilesystemList
from CedarBackup3.util import encodePath, executeCommand, isRunningAsRoot, pathJoin, resolveCommand, splitCommandLine

########################################################################
# Module-wide constants and variables
########################################################################

logger = logging.getLogger("CedarBackup3.log.peer")

DEF_RCP_COMMAND = ["/usr/bin/scp", "-B", "-q", "-C"]
DEF_RSH_COMMAND = ["/usr/bin/ssh"]
DEF_CBACK_COMMAND = "/usr/bin/cback3"

DEF_COLLECT_INDICATOR = "cback.collect"
DEF_STAGE_INDICATOR = "cback.stage"

SU_COMMAND = ["su"]


########################################################################
# LocalPeer class definition
########################################################################


class LocalPeer(object):
    ######################
    # Class documentation
    ######################

    """
    Backup peer representing a local peer in a backup pool.

    This is a class representing a local (non-network) peer in a backup pool.
    Local peers are backed up by simple filesystem copy operations.  A local
    peer has associated with it a name (typically, but not necessarily, a
    hostname) and a collect directory.

    The public methods other than the constructor are part of a "backup peer"
    interface shared with the ``RemotePeer`` class.

    """

    ##############
    # Constructor
    ##############

    def __init__(self, name, collectDir, ignoreFailureMode=None):
        """
        Initializes a local backup peer.

        Note that the collect directory must be an absolute path, but does not
        have to exist when the object is instantiated.  We do a lazy validation
        on this value since we could (potentially) be creating peer objects
        before an ongoing backup completed.

        Args:
           name: Name of the backup peer
           collectDir: Path to the peer's collect directory
           ignoreFailureMode: Ignore failure mode for this peer, one of ``VALID_FAILURE_MODES``
        Raises:
           ValueError: If the name is empty
           ValueError: If collect directory is not an absolute path
        """
        self._name = None
        self._collectDir = None
        self._ignoreFailureMode = None
        self.name = name
        self.collectDir = collectDir
        self.ignoreFailureMode = ignoreFailureMode

    #############
    # Properties
    #############

    def _setName(self, value):
        """
        Property target used to set the peer name.
        The value must be a non-empty string and cannot be ``None``.
        Raises:
           ValueError: If the value is an empty string or ``None``
        """
        if value is None or len(value) < 1:
            raise ValueError("Peer name must be a non-empty string.")
        self._name = value

    def _getName(self):
        """
        Property target used to get the peer name.
        """
        return self._name

    def _setCollectDir(self, value):
        """
        Property target used to set the collect directory.
        The value must be an absolute path and cannot be ``None``.
        It does not have to exist on disk at the time of assignment.
        Raises:
           ValueError: If the value is ``None`` or is not an absolute path
           ValueError: If a path cannot be encoded properly
        """
        if value is None or not (
            os.path.isabs(value) or posixpath.isabs(value)
        ):  # Python 3.13+ does not treat / as absolute on Windows
            raise ValueError("Collect directory must be an absolute path.")
        self._collectDir = encodePath(value)

    def _getCollectDir(self):
        """
        Property target used to get the collect directory.
        """
        return self._collectDir

    def _setIgnoreFailureMode(self, value):
        """
        Property target used to set the ignoreFailure mode.
        If not ``None``, the mode must be one of the values in :any:`VALID_FAILURE_MODES`.
        Raises:
           ValueError: If the value is not valid
        """
        if value is not None:
            if value not in VALID_FAILURE_MODES:
                raise ValueError("Ignore failure mode must be one of %s." % VALID_FAILURE_MODES)
        self._ignoreFailureMode = value

    def _getIgnoreFailureMode(self):
        """
        Property target used to get the ignoreFailure mode.
        """
        return self._ignoreFailureMode

    name = property(_getName, _setName, None, "Name of the peer.")
    collectDir = property(_getCollectDir, _setCollectDir, None, "Path to the peer's collect directory (an absolute local path).")
    ignoreFailureMode = property(_getIgnoreFailureMode, _setIgnoreFailureMode, None, "Ignore failure mode for peer.")

    #################
    # Public methods
    #################

    def stagePeer(self, targetDir, ownership=None, permissions=None):
        """
        Stages data from the peer into the indicated local target directory.

        The collect and target directories must both already exist before this
        method is called.  If passed in, ownership and permissions will be
        applied to the files that are copied.

        *Note:* The caller is responsible for checking that the indicator exists,
        if they care.  This function only stages the files within the directory.

        *Note:* If you have user/group as strings, call the :any:`util.getUidGid` function
        to get the associated uid/gid as an ownership tuple.

        Args:
           targetDir: Target directory to write data into
           ownership: Owner and group that files should have, tuple of numeric ``(uid, gid)``
           permissions: Unix permissions mode that the staged files should have, in octal like ``0640``
        Returns:
            Number of files copied from the source directory to the target directory

        Raises:
           ValueError: If collect directory is not a directory or does not exist
           ValueError: If target directory is not a directory, does not exist or is not absolute
           ValueError: If a path cannot be encoded properly
           IOError: If there were no files to stage (i.e. the directory was empty)
           IOError: If there is an IO error copying a file
           OSError: If there is an OS error copying or changing permissions on a file
        """
        targetDir = encodePath(targetDir)
        if not (os.path.isabs(targetDir) or posixpath.isabs(targetDir)):  # Python 3.13+ does not treat / as absolute on Windows
            logger.debug("Target directory [%s] not an absolute path.", targetDir)
            raise ValueError("Target directory must be an absolute path.")
        if not os.path.exists(self.collectDir) or not os.path.isdir(self.collectDir):
            logger.debug("Collect directory [%s] is not a directory or does not exist on disk.", self.collectDir)
            raise ValueError("Collect directory is not a directory or does not exist on disk.")
        if not os.path.exists(targetDir) or not os.path.isdir(targetDir):
            logger.debug("Target directory [%s] is not a directory or does not exist on disk.", targetDir)
            raise ValueError("Target directory is not a directory or does not exist on disk.")
        count = LocalPeer._copyLocalDir(self.collectDir, targetDir, ownership, permissions)
        if count == 0:
            raise IOError("Did not copy any files from local peer.")
        return count

    def checkCollectIndicator(self, collectIndicator=None):
        """
        Checks the collect indicator in the peer's staging directory.

        When a peer has completed collecting its backup files, it will write an
        empty indicator file into its collect directory.  This method checks to
        see whether that indicator has been written.  We're "stupid" here - if
        the collect directory doesn't exist, you'll naturally get back ``False``.

        If you need to, you can override the name of the collect indicator file
        by passing in a different name.

        Args:
           collectIndicator: Name of the collect indicator file to check
        Returns:
            Boolean true/false depending on whether the indicator exists
        Raises:
           ValueError: If a path cannot be encoded properly
        """
        collectIndicator = encodePath(collectIndicator)
        if collectIndicator is None:
            return os.path.exists(pathJoin(self.collectDir, DEF_COLLECT_INDICATOR))
        else:
            return os.path.exists(pathJoin(self.collectDir, collectIndicator))

    def writeStageIndicator(self, stageIndicator=None, ownership=None, permissions=None):
        """
        Writes the stage indicator in the peer's staging directory.

        When the master has completed collecting its backup files, it will write
        an empty indicator file into the peer's collect directory.  The presence
        of this file implies that the staging process is complete.

        If you need to, you can override the name of the stage indicator file by
        passing in a different name.

        *Note:* If you have user/group as strings, call the :any:`util.getUidGid`
        function to get the associated uid/gid as an ownership tuple.

        Args:
           stageIndicator: Name of the indicator file to write
           ownership: Owner and group that files should have, tuple of numeric ``(uid, gid)``
           permissions: Unix permissions mode that the staged files should have, in octal like ``0640``
        Raises:
           ValueError: If collect directory is not a directory or does not exist
           ValueError: If a path cannot be encoded properly
           IOError: If there is an IO error creating the file
           OSError: If there is an OS error creating or changing permissions on the file
        """
        stageIndicator = encodePath(stageIndicator)
        if not os.path.exists(self.collectDir) or not os.path.isdir(self.collectDir):
            logger.debug("Collect directory [%s] is not a directory or does not exist on disk.", self.collectDir)
            raise ValueError("Collect directory is not a directory or does not exist on disk.")
        if stageIndicator is None:
            fileName = pathJoin(self.collectDir, DEF_STAGE_INDICATOR)
        else:
            fileName = pathJoin(self.collectDir, stageIndicator)
        LocalPeer._copyLocalFile(None, fileName, ownership, permissions)  # None for sourceFile results in an empty target

    ##################
    # Private methods
    ##################

    @staticmethod
    def _copyLocalDir(sourceDir, targetDir, ownership=None, permissions=None):
        """
        Copies files from the source directory to the target directory.

        This function is not recursive.  Only the files in the directory will be
        copied.   Ownership and permissions will be left at their default values
        if new values are not specified.  The source and target directories are
        allowed to be soft links to a directory, but besides that soft links are
        ignored.

        *Note:* If you have user/group as strings, call the :any:`util.getUidGid`
        function to get the associated uid/gid as an ownership tuple.

        Args:
           sourceDir: Source directory
           targetDir: Target directory
           ownership: Owner and group that files should have, tuple of numeric ``(uid, gid)``
           permissions: Unix permissions mode that the staged files should have, in octal like ``0640``
        Returns:
            Number of files copied from the source directory to the target directory

        Raises:
           ValueError: If source or target is not a directory or does not exist
           ValueError: If a path cannot be encoded properly
           IOError: If there is an IO error copying the files
           OSError: If there is an OS error copying or changing permissions on a files
        """
        filesCopied = 0
        sourceDir = encodePath(sourceDir)
        targetDir = encodePath(targetDir)
        for fileName in os.listdir(sourceDir):
            sourceFile = pathJoin(sourceDir, fileName)
            targetFile = pathJoin(targetDir, fileName)
            LocalPeer._copyLocalFile(sourceFile, targetFile, ownership, permissions)
            filesCopied += 1
        return filesCopied

    @staticmethod
    def _copyLocalFile(sourceFile=None, targetFile=None, ownership=None, permissions=None, overwrite=True):
        """
        Copies a source file to a target file.

        If the source file is ``None`` then the target file will be created or
        overwritten as an empty file.  If the target file is ``None``, this method
        is a no-op.  Attempting to copy a soft link or a directory will result in
        an exception.

        *Note:* If you have user/group as strings, call the :any:`util.getUidGid`
        function to get the associated uid/gid as an ownership tuple.

        *Note:* We will not overwrite a target file that exists when this method
        is invoked.  If the target already exists, we'll raise an exception.

        Args:
           sourceFile: Source file to copy
           targetFile: Target file to create
           ownership: Owner and group that files should have, tuple of numeric ``(uid, gid)``
           permissions: Unix permissions mode that the staged files should have, in octal like ``0640``
           overwrite: Indicates whether it's OK to overwrite the target file
        Raises:
           ValueError: If the passed-in source file is not a regular file
           ValueError: If a path cannot be encoded properly
           IOError: If the target file already exists
           IOError: If there is an IO error copying the file
           OSError: If there is an OS error copying or changing permissions on a file
        """
        targetFile = encodePath(targetFile)
        sourceFile = encodePath(sourceFile)
        if targetFile is None:
            return
        if not overwrite:
            if os.path.exists(targetFile):
                raise IOError("Target file [%s] already exists." % targetFile)
        if sourceFile is None:
            with open(targetFile, "w") as f:  # pylint: disable=unspecified-encoding
                f.write("")
        else:
            if os.path.isfile(sourceFile) and not os.path.islink(sourceFile):
                shutil.copy(sourceFile, targetFile)
            else:
                logger.debug("Source [%s] is not a regular file.", sourceFile)
                raise ValueError("Source is not a regular file.")
        if ownership is not None:
            if sys.platform != "win32":
                os.chown(targetFile, ownership[0], ownership[1])  # pylint: disable=no-member
        if permissions is not None:
            os.chmod(targetFile, permissions)


########################################################################
# RemotePeer class definition
########################################################################


class RemotePeer(object):
    ######################
    # Class documentation
    ######################

    """
    Backup peer representing a remote peer in a backup pool.

    This is a class representing a remote (networked) peer in a backup pool.
    Remote peers are backed up using an rcp-compatible copy command.  A remote
    peer has associated with it a name (which must be a valid hostname), a
    collect directory, a working directory and a copy method (an rcp-compatible
    command).

    You can also set an optional local user value.  This username will be used
    as the local user for any remote copies that are required.  It can only be
    used if the root user is executing the backup.  The root user will ``su`` to
    the local user and execute the remote copies as that user.

    The copy method is associated with the peer and not with the actual request
    to copy, because we can envision that each remote host might have a
    different connect method.

    The public methods other than the constructor are part of a "backup peer"
    interface shared with the ``LocalPeer`` class.

    """

    ##############
    # Constructor
    ##############

    def __init__(
        self,
        name=None,
        collectDir=None,
        workingDir=None,
        remoteUser=None,
        rcpCommand=None,
        localUser=None,
        rshCommand=None,
        cbackCommand=None,
        ignoreFailureMode=None,
    ):
        """
        Initializes a remote backup peer.

        *Note:* If provided, each command will eventually be parsed into a list of
        strings suitable for passing to ``util.executeCommand`` in order to avoid
        security holes related to shell interpolation.   This parsing will be
        done by the :any:`util.splitCommandLine` function.  See the documentation for
        that function for some important notes about its limitations.

        Args:
           name: Name of the backup peer, a valid DNS name
           collectDir: Path to the peer's collect directory, absolute path
           workingDir: Working directory that can be used to create temporary files, etc, an absolute path
           remoteUser: Name of the Cedar Backup user on the remote peer
           localUser: Name of the Cedar Backup user on the current host
           rcpCommand: An rcp-compatible copy command to use for copying files from the peer
           rshCommand: An rsh-compatible copy command to use for remote shells to the peer
           cbackCommand: A chack-compatible command to use for executing managed actions
           ignoreFailureMode: Ignore failure mode for this peer, one of ``VALID_FAILURE_MODES``
        Raises:
           ValueError: If collect directory is not an absolute path
        """
        self._name = None
        self._collectDir = None
        self._workingDir = None
        self._remoteUser = None
        self._localUser = None
        self._rcpCommand = None
        self._rcpCommandList = None
        self._rshCommand = None
        self._rshCommandList = None
        self._cbackCommand = None
        self._ignoreFailureMode = None
        self.name = name
        self.collectDir = collectDir
        self.workingDir = workingDir
        self.remoteUser = remoteUser
        self.localUser = localUser
        self.rcpCommand = rcpCommand
        self.rshCommand = rshCommand
        self.cbackCommand = cbackCommand
        self.ignoreFailureMode = ignoreFailureMode

    #############
    # Properties
    #############

    def _setName(self, value):
        """
        Property target used to set the peer name.
        The value must be a non-empty string and cannot be ``None``.
        Raises:
           ValueError: If the value is an empty string or ``None``
        """
        if value is None or len(value) < 1:
            raise ValueError("Peer name must be a non-empty string.")
        self._name = value

    def _getName(self):
        """
        Property target used to get the peer name.
        """
        return self._name

    def _setCollectDir(self, value):
        """
        Property target used to set the collect directory.
        The value must be an absolute path and cannot be ``None``.
        It does not have to exist on disk at the time of assignment.
        Raises:
           ValueError: If the value is ``None`` or is not an absolute path
           ValueError: If the value cannot be encoded properly
        """
        if value is not None:
            if not (os.path.isabs(value) or posixpath.isabs(value)):  # Python 3.13+ does not treat / as absolute on Windows
                raise ValueError("Collect directory must be an absolute path.")
        self._collectDir = encodePath(value)

    def _getCollectDir(self):
        """
        Property target used to get the collect directory.
        """
        return self._collectDir

    def _setWorkingDir(self, value):
        """
        Property target used to set the working directory.
        The value must be an absolute path and cannot be ``None``.
        Raises:
           ValueError: If the value is ``None`` or is not an absolute path
           ValueError: If the value cannot be encoded properly
        """
        if value is not None:
            if not (os.path.isabs(value) or posixpath.isabs(value)):  # Python 3.13+ does not treat / as absolute on Windows
                raise ValueError("Working directory must be an absolute path.")
        self._workingDir = encodePath(value)

    def _getWorkingDir(self):
        """
        Property target used to get the working directory.
        """
        return self._workingDir

    def _setRemoteUser(self, value):
        """
        Property target used to set the remote user.
        The value must be a non-empty string and cannot be ``None``.
        Raises:
           ValueError: If the value is an empty string or ``None``
        """
        if value is None or len(value) < 1:
            raise ValueError("Peer remote user must be a non-empty string.")
        self._remoteUser = value

    def _getRemoteUser(self):
        """
        Property target used to get the remote user.
        """
        return self._remoteUser

    def _setLocalUser(self, value):
        """
        Property target used to set the local user.
        The value must be a non-empty string if it is not ``None``.
        Raises:
           ValueError: If the value is an empty string
        """
        if value is not None:
            if len(value) < 1:
                raise ValueError("Peer local user must be a non-empty string.")
        self._localUser = value

    def _getLocalUser(self):
        """
        Property target used to get the local user.
        """
        return self._localUser

    def _setRcpCommand(self, value):
        """
        Property target to set the rcp command.

        The value must be a non-empty string or ``None``.  Its value is stored in
        the two forms: "raw" as provided by the client, and "parsed" into a list
        suitable for being passed to :any:`util.executeCommand` via
        :any:`util.splitCommandLine`.

        However, all the caller will ever see via the property is the actual
        value they set (which includes seeing ``None``, even if we translate that
        internally to ``DEF_RCP_COMMAND``).  Internally, we should always use
        ``self._rcpCommandList`` if we want the actual command list.

        Raises:
           ValueError: If the value is an empty string
        """
        if value is None:
            self._rcpCommand = None
            self._rcpCommandList = DEF_RCP_COMMAND
        else:
            if len(value) >= 1:
                self._rcpCommand = value
                self._rcpCommandList = splitCommandLine(self._rcpCommand)
            else:
                raise ValueError("The rcp command must be a non-empty string.")

    def _getRcpCommand(self):
        """
        Property target used to get the rcp command.
        """
        return self._rcpCommand

    def _setRshCommand(self, value):
        """
        Property target to set the rsh command.

        The value must be a non-empty string or ``None``.  Its value is stored in
        the two forms: "raw" as provided by the client, and "parsed" into a list
        suitable for being passed to :any:`util.executeCommand` via
        :any:`util.splitCommandLine`.

        However, all the caller will ever see via the property is the actual
        value they set (which includes seeing ``None``, even if we translate that
        internally to ``DEF_RSH_COMMAND``).  Internally, we should always use
        ``self._rshCommandList`` if we want the actual command list.

        Raises:
           ValueError: If the value is an empty string
        """
        if value is None:
            self._rshCommand = None
            self._rshCommandList = DEF_RSH_COMMAND
        else:
            if len(value) >= 1:
                self._rshCommand = value
                self._rshCommandList = splitCommandLine(self._rshCommand)
            else:
                raise ValueError("The rsh command must be a non-empty string.")

    def _getRshCommand(self):
        """
        Property target used to get the rsh command.
        """
        return self._rshCommand

    def _setCbackCommand(self, value):
        """
        Property target to set the cback command.

        The value must be a non-empty string or ``None``.  Unlike the other
        command, this value is only stored in the "raw" form provided by the
        client.

        Raises:
           ValueError: If the value is an empty string
        """
        if value is None:
            self._cbackCommand = None
        else:
            if len(value) >= 1:
                self._cbackCommand = value
            else:
                raise ValueError("The cback command must be a non-empty string.")

    def _getCbackCommand(self):
        """
        Property target used to get the cback command.
        """
        return self._cbackCommand

    def _setIgnoreFailureMode(self, value):
        """
        Property target used to set the ignoreFailure mode.
        If not ``None``, the mode must be one of the values in :any:`VALID_FAILURE_MODES`.
        Raises:
           ValueError: If the value is not valid
        """
        if value is not None:
            if value not in VALID_FAILURE_MODES:
                raise ValueError("Ignore failure mode must be one of %s." % VALID_FAILURE_MODES)
        self._ignoreFailureMode = value

    def _getIgnoreFailureMode(self):
        """
        Property target used to get the ignoreFailure mode.
        """
        return self._ignoreFailureMode

    name = property(_getName, _setName, None, "Name of the peer (a valid DNS hostname).")
    collectDir = property(_getCollectDir, _setCollectDir, None, "Path to the peer's collect directory (an absolute local path).")
    workingDir = property(_getWorkingDir, _setWorkingDir, None, "Path to the peer's working directory (an absolute local path).")
    remoteUser = property(_getRemoteUser, _setRemoteUser, None, "Name of the Cedar Backup user on the remote peer.")
    localUser = property(_getLocalUser, _setLocalUser, None, "Name of the Cedar Backup user on the current host.")
    rcpCommand = property(_getRcpCommand, _setRcpCommand, None, "An rcp-compatible copy command to use for copying files.")
    rshCommand = property(_getRshCommand, _setRshCommand, None, "An rsh-compatible command to use for remote shells to the peer.")
    cbackCommand = property(
        _getCbackCommand, _setCbackCommand, None, "A chack-compatible command to use for executing managed actions."
    )
    ignoreFailureMode = property(_getIgnoreFailureMode, _setIgnoreFailureMode, None, "Ignore failure mode for peer.")

    #################
    # Public methods
    #################

    def stagePeer(self, targetDir, ownership=None, permissions=None):
        """
        Stages data from the peer into the indicated local target directory.

        The target directory must already exist before this method is called.  If
        passed in, ownership and permissions will be applied to the files that
        are copied.

        *Note:* The returned count of copied files might be inaccurate if some of
        the copied files already existed in the staging directory prior to the
        copy taking place.  We don't clear the staging directory first, because
        some extension might also be using it.

        *Note:* If you have user/group as strings, call the :any:`util.getUidGid` function
        to get the associated uid/gid as an ownership tuple.

        *Note:* Unlike the local peer version of this method, an I/O error might
        or might not be raised if the directory is empty.  Since we're using a
        remote copy method, we just don't have the fine-grained control over our
        exceptions that's available when we can look directly at the filesystem,
        and we can't control whether the remote copy method thinks an empty
        directory is an error.

        Args:
           targetDir: Target directory to write data into
           ownership: Owner and group that files should have, tuple of numeric ``(uid, gid)``
           permissions: Unix permissions mode that the staged files should have, in octal like ``0640``
        Returns:
            Number of files copied from the source directory to the target directory

        Raises:
           ValueError: If target directory is not a directory, does not exist or is not absolute
           ValueError: If a path cannot be encoded properly
           IOError: If there were no files to stage (i.e. the directory was empty)
           IOError: If there is an IO error copying a file
           OSError: If there is an OS error copying or changing permissions on a file
        """
        targetDir = encodePath(targetDir)
        if not (os.path.isabs(targetDir) or posixpath.isabs(targetDir)):  # Python 3.13+ does not treat / as absolute on Windows
            logger.debug("Target directory [%s] not an absolute path.", targetDir)
            raise ValueError("Target directory must be an absolute path.")
        if not os.path.exists(targetDir) or not os.path.isdir(targetDir):
            logger.debug("Target directory [%s] is not a directory or does not exist on disk.", targetDir)
            raise ValueError("Target directory is not a directory or does not exist on disk.")
        count = RemotePeer._copyRemoteDir(
            self.remoteUser,
            self.localUser,
            self.name,
            self._rcpCommand,
            self._rcpCommandList,
            self.collectDir,
            targetDir,
            ownership,
            permissions,
        )
        if count == 0:
            raise IOError("Did not copy any files from local peer.")
        return count

    def checkCollectIndicator(self, collectIndicator=None):
        """
        Checks the collect indicator in the peer's staging directory.

        When a peer has completed collecting its backup files, it will write an
        empty indicator file into its collect directory.  This method checks to
        see whether that indicator has been written.  If the remote copy command
        fails, we return ``False`` as if the file weren't there.

        If you need to, you can override the name of the collect indicator file
        by passing in a different name.

        *Note:* Apparently, we can't count on all rcp-compatible implementations
        to return sensible errors for some error conditions.  As an example, the
        ``scp`` command in Debian 'woody' returns a zero (normal) status even when
        it can't find a host or if the login or path is invalid.  Because of
        this, the implementation of this method is rather convoluted.

        Args:
           collectIndicator: Name of the collect indicator file to check
        Returns:
            Boolean true/false depending on whether the indicator exists
        Raises:
           ValueError: If a path cannot be encoded properly
        """
        try:
            if collectIndicator is None:
                sourceFile = pathJoin(self.collectDir, DEF_COLLECT_INDICATOR)
                targetFile = pathJoin(self.workingDir, DEF_COLLECT_INDICATOR)
            else:
                collectIndicator = encodePath(collectIndicator)
                sourceFile = pathJoin(self.collectDir, collectIndicator)
                targetFile = pathJoin(self.workingDir, collectIndicator)
            logger.debug("Fetch remote [%s] into [%s].", sourceFile, targetFile)
            if os.path.exists(targetFile):
                try:
                    os.remove(targetFile)
                except:
                    raise Exception("Error: collect indicator [%s] already exists!" % targetFile)
            try:
                RemotePeer._copyRemoteFile(
                    self.remoteUser,
                    self.localUser,
                    self.name,
                    self._rcpCommand,
                    self._rcpCommandList,
                    sourceFile,
                    targetFile,
                    overwrite=False,
                )
                if os.path.exists(targetFile):
                    return True
                else:
                    return False
            except Exception as e:
                logger.info("Failed looking for collect indicator: %s", e)
                return False
        finally:
            if os.path.exists(targetFile):
                try:
                    os.remove(targetFile)
                except:
                    pass

    def writeStageIndicator(self, stageIndicator=None):
        """
        Writes the stage indicator in the peer's staging directory.

        When the master has completed collecting its backup files, it will write
        an empty indicator file into the peer's collect directory.  The presence
        of this file implies that the staging process is complete.

        If you need to, you can override the name of the stage indicator file by
        passing in a different name.

        *Note:* If you have user/group as strings, call the :any:`util.getUidGid` function
        to get the associated uid/gid as an ownership tuple.

        Args:
           stageIndicator: Name of the indicator file to write
        Raises:
           ValueError: If a path cannot be encoded properly
           IOError: If there is an IO error creating the file
           OSError: If there is an OS error creating or changing permissions on the file
        """
        stageIndicator = encodePath(stageIndicator)
        if stageIndicator is None:
            sourceFile = pathJoin(self.workingDir, DEF_STAGE_INDICATOR)
            targetFile = pathJoin(self.collectDir, DEF_STAGE_INDICATOR)
        else:
            sourceFile = pathJoin(self.workingDir, DEF_STAGE_INDICATOR)
            targetFile = pathJoin(self.collectDir, stageIndicator)
        try:
            if not os.path.exists(sourceFile):
                with open(sourceFile, "w") as f:  # pylint: disable=unspecified-encoding
                    f.write("")
            RemotePeer._pushLocalFile(
                self.remoteUser, self.localUser, self.name, self._rcpCommand, self._rcpCommandList, sourceFile, targetFile
            )
        finally:
            if os.path.exists(sourceFile):
                try:
                    os.remove(sourceFile)
                except:
                    pass

    def executeRemoteCommand(self, command):
        """
        Executes a command on the peer via remote shell.

        Args:
           command: Command to execute
        Raises:
           IOError: If there is an error executing the command on the remote peer
        """
        RemotePeer._executeRemoteCommand(
            self.remoteUser, self.localUser, self.name, self._rshCommand, self._rshCommandList, command
        )

    def executeManagedAction(self, action, fullBackup):
        """
        Executes a managed action on this peer.

        Args:
           action: Name of the action to execute
           fullBackup: Whether a full backup should be executed

        Raises:
           IOError: If there is an error executing the action on the remote peer
        """
        try:
            command = RemotePeer._buildCbackCommand(self.cbackCommand, action, fullBackup)
            self.executeRemoteCommand(command)
        except IOError as e:
            logger.info(e)
            raise IOError("Failed to execute action [%s] on managed client [%s]." % (action, self.name))

    ##################
    # Private methods
    ##################

    @staticmethod
    def _getDirContents(path):
        """
        Returns the contents of a directory in terms of a Set.

        The directory's contents are read as a :any:`FilesystemList` containing only
        files, and then the list is converted into a set object for later use.

        Args:
           path: Directory path to get contents for
        Returns:
            Set of files in the directory
        Raises:
           ValueError: If path is not a directory or does not exist
        """
        contents = FilesystemList()
        contents.excludeDirs = True
        contents.excludeLinks = True
        contents.addDirContents(path)
        return set(contents)

    @staticmethod
    def _copyRemoteDir(
        remoteUser, localUser, remoteHost, rcpCommand, rcpCommandList, sourceDir, targetDir, ownership=None, permissions=None
    ):
        """
        Copies files from the source directory to the target directory.

        This function is not recursive.  Only the files in the directory will be
        copied.   Ownership and permissions will be left at their default values
        if new values are not specified.  Behavior when copying soft links from
        the collect directory is dependent on the behavior of the specified rcp
        command.

        *Note:* The returned count of copied files might be inaccurate if some of
        the copied files already existed in the staging directory prior to the
        copy taking place.  We don't clear the staging directory first, because
        some extension might also be using it.

        *Note:* If you have user/group as strings, call the :any:`util.getUidGid` function
        to get the associated uid/gid as an ownership tuple.

        *Note:* We don't have a good way of knowing exactly what files we copied
        down from the remote peer, unless we want to parse the output of the rcp
        command (ugh).  We could change permissions on everything in the target
        directory, but that's kind of ugly too.  Instead, we use Python's set
        functionality to figure out what files were added while we executed the
        rcp command.  This isn't perfect - for instance, it's not correct if
        someone else is messing with the directory at the same time we're doing
        the remote copy - but it's about as good as we're going to get.

        *Note:* Apparently, we can't count on all rcp-compatible implementations
        to return sensible errors for some error conditions.  As an example, the
        ``scp`` command in Debian 'woody' returns a zero (normal) status even
        when it can't find a host or if the login or path is invalid.  We try
        to work around this by issuing ``IOError`` if we don't copy any files from
        the remote host.

        Args:
           remoteUser: Name of the Cedar Backup user on the remote peer
           localUser: Name of the Cedar Backup user on the current host
           remoteHost: Hostname of the remote peer
           rcpCommand: An rcp-compatible copy command to use for copying files from the peer
           rcpCommandList: An rcp-compatible copy command to use for copying files, as for :any:`util.executeCommand`
           sourceDir: Source directory
           targetDir: Target directory
           ownership: Owner and group that files should have, tuple of numeric ``(uid, gid)``
           permissions: Unix permissions mode that the staged files should have, in octal like ``0640``
        Returns:
            Number of files copied from the source directory to the target directory

        Raises:
           ValueError: If source or target is not a directory or does not exist
           IOError: If there is an IO error copying the files
        """
        beforeSet = RemotePeer._getDirContents(targetDir)
        if localUser is not None:
            try:
                if not isRunningAsRoot():
                    raise IOError("Only root can remote copy as another user.")
            except AttributeError:
                pass
            actualCommand = "%s %s@%s:%s/* %s" % (rcpCommand, remoteUser, remoteHost, sourceDir, targetDir)
            command = resolveCommand(SU_COMMAND)
            result = executeCommand(command, [localUser, "-c", actualCommand])[0]
            if result != 0:
                raise IOError("Error (%d) copying files from remote host as local user [%s]." % (result, localUser))
        else:
            copySource = "%s@%s:%s/*" % (remoteUser, remoteHost, sourceDir)
            command = resolveCommand(rcpCommandList)
            result = executeCommand(command, [copySource, targetDir])[0]
            if result != 0:
                raise IOError("Error (%d) copying files from remote host." % result)
        afterSet = RemotePeer._getDirContents(targetDir)
        if len(afterSet) == 0:
            raise IOError("Did not copy any files from remote peer.")
        differenceSet = afterSet.difference(beforeSet)  # files we added as part of copy
        if len(differenceSet) == 0:
            raise IOError("Apparently did not copy any new files from remote peer.")
        for targetFile in differenceSet:
            if ownership is not None:
                if sys.platform != "win32":
                    os.chown(targetFile, ownership[0], ownership[1])  # pylint: disable=no-member
            if permissions is not None:
                os.chmod(targetFile, permissions)
        return len(differenceSet)

    @staticmethod
    def _copyRemoteFile(
        remoteUser,
        localUser,
        remoteHost,
        rcpCommand,
        rcpCommandList,
        sourceFile,
        targetFile,
        ownership=None,
        permissions=None,
        overwrite=True,
    ):
        """
        Copies a remote source file to a target file.

        *Note:* Internally, we have to go through and escape any spaces in the
        source path with double-backslash, otherwise things get screwed up.   It
        doesn't seem to be required in the target path. I hope this is portable
        to various different rcp methods, but I guess it might not be (all I have
        to test with is OpenSSH).

        *Note:* If you have user/group as strings, call the :any:`util.getUidGid` function
        to get the associated uid/gid as an ownership tuple.

        *Note:* We will not overwrite a target file that exists when this method
        is invoked.  If the target already exists, we'll raise an exception.

        *Note:* Apparently, we can't count on all rcp-compatible implementations
        to return sensible errors for some error conditions.  As an example, the
        ``scp`` command in Debian 'woody' returns a zero (normal) status even when
        it can't find a host or if the login or path is invalid.  We try to work
        around this by issuing ``IOError`` the target file does not exist when
        we're done.

        Args:
           remoteUser: Name of the Cedar Backup user on the remote peer
           remoteHost: Hostname of the remote peer
           localUser: Name of the Cedar Backup user on the current host
           rcpCommand: An rcp-compatible copy command to use for copying files from the peer
           rcpCommandList: An rcp-compatible copy command to use for copying files, as for :any:`util.executeCommand`
           sourceFile: Source file to copy
           targetFile: Target file to create
           ownership: Owner and group that files should have, tuple of numeric ``(uid, gid)``
           permissions: Unix permissions mode that the staged files should have, in octal like ``0640``
           overwrite: Indicates whether it's OK to overwrite the target file
        Raises:
           IOError: If the target file already exists
           IOError: If there is an IO error copying the file
           OSError: If there is an OS error changing permissions on the file
        """
        if not overwrite:
            if os.path.exists(targetFile):
                raise IOError("Target file [%s] already exists." % targetFile)
        if localUser is not None:
            try:
                if not isRunningAsRoot():
                    raise IOError("Only root can remote copy as another user.")
            except AttributeError:
                pass
            actualCommand = "%s %s@%s:%s %s" % (rcpCommand, remoteUser, remoteHost, sourceFile.replace(" ", "\\ "), targetFile)
            command = resolveCommand(SU_COMMAND)
            result = executeCommand(command, [localUser, "-c", actualCommand])[0]
            if result != 0:
                raise IOError("Error (%d) copying [%s] from remote host as local user [%s]." % (result, sourceFile, localUser))
        else:
            copySource = "%s@%s:%s" % (remoteUser, remoteHost, sourceFile.replace(" ", "\\ "))
            command = resolveCommand(rcpCommandList)
            result = executeCommand(command, [copySource, targetFile])[0]
            if result != 0:
                raise IOError("Error (%d) copying [%s] from remote host." % (result, sourceFile))
        if not os.path.exists(targetFile):
            raise IOError("Apparently unable to copy file from remote host.")
        if ownership is not None:
            if sys.platform != "win32":
                os.chown(targetFile, ownership[0], ownership[1])  # pylint: disable=no-member
        if permissions is not None:
            os.chmod(targetFile, permissions)

    @staticmethod
    def _pushLocalFile(remoteUser, localUser, remoteHost, rcpCommand, rcpCommandList, sourceFile, targetFile, overwrite=True):
        """
        Copies a local source file to a remote host.

        *Note:* We will not overwrite a target file that exists when this method
        is invoked.  If the target already exists, we'll raise an exception.

        *Note:* Internally, we have to go through and escape any spaces in the
        source and target paths with double-backslash, otherwise things get
        screwed up.  I hope this is portable to various different rcp methods,
        but I guess it might not be (all I have to test with is OpenSSH).

        *Note:* If you have user/group as strings, call the :any:`util.getUidGid` function
        to get the associated uid/gid as an ownership tuple.

        Args:
           remoteUser: Name of the Cedar Backup user on the remote peer
           localUser: Name of the Cedar Backup user on the current host
           remoteHost: Hostname of the remote peer
           rcpCommand: An rcp-compatible copy command to use for copying files from the peer
           rcpCommandList: An rcp-compatible copy command to use for copying files, as for :any:`util.executeCommand`
           sourceFile: Source file to copy
           targetFile: Target file to create
           overwrite: Indicates whether it's OK to overwrite the target file
        Raises:
           IOError: If there is an IO error copying the file
           OSError: If there is an OS error changing permissions on the file
        """
        if not overwrite:
            if os.path.exists(targetFile):
                raise IOError("Target file [%s] already exists." % targetFile)
        if localUser is not None:
            try:
                if not isRunningAsRoot():
                    raise IOError("Only root can remote copy as another user.")
            except AttributeError:
                pass
            actualCommand = '%s "%s" "%s@%s:%s"' % (rcpCommand, sourceFile, remoteUser, remoteHost, targetFile)
            command = resolveCommand(SU_COMMAND)
            result = executeCommand(command, [localUser, "-c", actualCommand])[0]
            if result != 0:
                raise IOError("Error (%d) copying [%s] to remote host as local user [%s]." % (result, sourceFile, localUser))
        else:
            copyTarget = "%s@%s:%s" % (remoteUser, remoteHost, targetFile.replace(" ", "\\ "))
            command = resolveCommand(rcpCommandList)
            result = executeCommand(command, [sourceFile.replace(" ", "\\ "), copyTarget])[0]
            if result != 0:
                raise IOError("Error (%d) copying [%s] to remote host." % (result, sourceFile))

    @staticmethod
    def _executeRemoteCommand(remoteUser, localUser, remoteHost, rshCommand, rshCommandList, remoteCommand):
        """
        Executes a command on the peer via remote shell.

        Args:
           remoteUser: Name of the Cedar Backup user on the remote peer
           localUser: Name of the Cedar Backup user on the current host
           remoteHost: Hostname of the remote peer
           rshCommand: An rsh-compatible copy command to use for remote shells to the peer
           rshCommandList: An rsh-compatible copy command to use for remote shells to the peer, as for :any:`util.executeCommand`
           remoteCommand: The command to be executed on the remote host, with no special shell characters
        Raises:
           IOError: If there is an error executing the remote command
        """
        actualCommand = "%s %s@%s '%s'" % (rshCommand, remoteUser, remoteHost, remoteCommand)
        if localUser is not None:
            try:
                if not isRunningAsRoot():
                    raise IOError("Only root can remote shell as another user.")
            except AttributeError:
                pass
            command = resolveCommand(SU_COMMAND)
            result = executeCommand(command, [localUser, "-c", actualCommand])[0]
            if result != 0:
                raise IOError('Command failed [su -c %s "%s"]' % (localUser, actualCommand))
        else:
            command = resolveCommand(rshCommandList)
            result = executeCommand(command, ["%s@%s" % (remoteUser, remoteHost), "%s" % remoteCommand])[0]
            if result != 0:
                raise IOError("Command failed [%s]" % (actualCommand))

    @staticmethod
    def _buildCbackCommand(cbackCommand, action, fullBackup):
        """
        Builds a Cedar Backup command line for the named action.

        *Note:* If the cback command is None, then DEF_CBACK_COMMAND is used.

        Args:
           cbackCommand: cback command to execute, including required options
           action: Name of the action to execute
           fullBackup: Whether a full backup should be executed

        Returns:
            String suitable for passing to :any:`_executeRemoteCommand` as remoteCommand
        Raises:
           ValueError: If action is None
        """
        if action is None:
            raise ValueError("Action cannot be None.")
        if cbackCommand is None:
            cbackCommand = DEF_CBACK_COMMAND
        if fullBackup:
            return "%s --full %s" % (cbackCommand, action)
        else:
            return "%s %s" % (cbackCommand, action)



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
#              C E D A R
#          S O L U T I O N S       "Software done right."
#           S O F T W A R E
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Copyright (c) 2004-2007,2010,2015 Kenneth J. Pronovici.
# All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License,
# Version 2, as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# Copies of the GNU General Public License are available from
# the Free Software Foundation website, http://www.gnu.org/.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Author   : Kenneth J. Pronovici <pronovic@ieee.org>
# Language : Python 3
# Project  : Cedar Backup, release 3
# Purpose  : Tests CD writer functionality.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

########################################################################
# Module documentation
########################################################################

"""
Unit tests for CedarBackup3/writers/cdwriter.py.

This code was consolidated from writertests.py and imagetests.py at the same
time cdwriter.py was created.

Code Coverage
=============

   This module contains individual tests for the public classes implemented in
   cdwriter.py.

   Unfortunately, it's rather difficult to test this code in an automated
   fashion, even if you have access to a physical CD writer drive.  It's even
   more difficult to test it if you are running on some build daemon (think of
   a Debian autobuilder) which can't be expected to have any hardware or any
   media that you could write to.  Because of this, there aren't any tests
   below that actually cause CD media to be written to.

   As a compromise, much of the implementation is in terms of private static
   methods that have well-defined behaviors.  Normally, I prefer to only test
   the public interface to class, but in this case, testing the private methods
   will help give us some reasonable confidence in the code, even if we can't
   write a physical disc or can't run all of the tests.  This isn't perfect,
   but it's better than nothing.

Naming Conventions
==================

   I prefer to avoid large unit tests which validate more than one piece of
   functionality, and I prefer to avoid using overly descriptive (read: long)
   test names, as well.  Instead, I use lots of very small tests that each
   validate one specific thing.  These small tests are then named with an index
   number, yielding something like ``testAddDir_001`` or ``testValidate_010``.
   Each method has a docstring describing what it's supposed to accomplish.  I
   feel that this makes it easier to judge how important a given failure is,
   and also makes it somewhat easier to diagnose and fix individual problems.

Full vs. Reduced Tests
======================

   Some Cedar Backup regression tests require a specialized environment in
   order to run successfully.  This environment won't necessarily be available
   on every build system out there (for instance, on a Debian autobuilder).
   Because of this, the default behavior is to run a "reduced feature set" test
   suite that has no surprising system, kernel or network requirements.  There
   are no special dependencies for these tests.

   I used to try and run tests against an actual device, to make sure that this
   worked.  However, those tests ended up being kind of bogus, because my main
   development environment doesn't have a writer, and even if it had one, any
   device with the same name on another user's system wouldn't necessarily
   return sensible results.  That's just pointless.  We'll just have to rely on
   the other tests to make sure that things seem sensible.

@author Kenneth J. Pronovici <pronovic@ieee.org>
"""


########################################################################
# Import modules and do runtime validations
########################################################################

import unittest

from CedarBackup3.testutil import configureLogging
from CedarBackup3.writers.cdwriter import (
    MEDIA_CDR_74,
    MEDIA_CDR_80,
    MEDIA_CDRW_74,
    MEDIA_CDRW_80,
    CdWriter,
    MediaCapacity,
    MediaDefinition,
)

#######################################################################
# Module-wide configuration and constants
#######################################################################

MB650 = 650.0 * 1024.0 * 1024.0  # 650 MB
MB700 = 700.0 * 1024.0 * 1024.0  # 700 MB
ILEAD = 11400.0 * 2048.0  # Initial lead-in
SLEAD = 6900.0 * 2048.0  # Session lead-in

DATA_DIRS = [
    "./data",
    "./tests/data",
]
RESOURCES = [
    "tree9.tar.gz",
]

SUDO_CMD = [
    "sudo",
]
HDIUTIL_CMD = [
    "hdiutil",
]

INVALID_FILE = "bogus"  # This file name should never exist


#######################################################################
# Test Case Classes
#######################################################################

############################
# TestMediaDefinition class
############################


class TestMediaDefinition(unittest.TestCase):
    """Tests for the MediaDefinition class."""

    @classmethod
    def setUpClass(cls):
        configureLogging()

    def testConstructor_001(self):
        """
        Test the constructor with an invalid media type.
        """
        self.assertRaises(ValueError, MediaDefinition, 100)

    def testConstructor_002(self):
        """
        Test the constructor with the ``MEDIA_CDR_74`` media type.
        """
        media = MediaDefinition(MEDIA_CDR_74)
        self.assertEqual(MEDIA_CDR_74, media.mediaType)
        self.assertEqual(False, media.rewritable)
        self.assertNotEqual(0, media.initialLeadIn)  # just care that it's set, not what its value is
        self.assertNotEqual(0, media.leadIn)  # just care that it's set, not what its value is
        self.assertEqual(332800, media.capacity)

    def testConstructor_003(self):
        """
        Test the constructor with the ``MEDIA_CDRW_74`` media type.
        """
        media = MediaDefinition(MEDIA_CDRW_74)
        self.assertEqual(MEDIA_CDRW_74, media.mediaType)
        self.assertEqual(True, media.rewritable)
        self.assertNotEqual(0, media.initialLeadIn)  # just care that it's set, not what its value is
        self.assertNotEqual(0, media.leadIn)  # just care that it's set, not what its value is
        self.assertEqual(332800, media.capacity)

    def testConstructor_004(self):
        """
        Test the constructor with the ``MEDIA_CDR_80`` media type.
        """
        media = MediaDefinition(MEDIA_CDR_80)
        self.assertEqual(MEDIA_CDR_80, media.mediaType)
        self.assertEqual(False, media.rewritable)
        self.assertNotEqual(0, media.initialLeadIn)  # just care that it's set, not what its value is
        self.assertNotEqual(0, media.leadIn)  # just care that it's set, not what its value is
        self.assertEqual(358400, media.capacity)

    def testConstructor_005(self):
        """
        Test the constructor with the ``MEDIA_CDRW_80`` media type.
        """
        media = MediaDefinition(MEDIA_CDRW_80)
        self.assertEqual(MEDIA_CDRW_80, media.mediaType)
        self.assertEqual(True, media.rewritable)
        self.assertNotEqual(0, media.initialLeadIn)  # just care that it's set, not what its value is
        self.assertNotEqual(0, media.leadIn)  # just care that it's set, not what its value is
        self.assertEqual(358400, media.capacity)


############################
# TestMediaCapacity class
############################


class TestMediaCapacity(unittest.TestCase):
    """Tests for the MediaCapacity class."""

    @classmethod
    def setUpClass(cls):
        configureLogging()

    def testConstructor_001(self):
        """
        Test the constructor.
        """
        capacity = MediaCapacity(100, 200, (300, 400))
        self.assertEqual(100, capacity.bytesUsed)
        self.assertEqual(200, capacity.bytesAvailable)
        self.assertEqual((300, 400), capacity.boundaries)


#####################
# TestCdWriter class
#####################


class TestCdWriter(unittest.TestCase):
    """Tests for the CdWriter class."""

    ################
    # Setup methods
    ################

    @classmethod
    def setUpClass(cls):
        configureLogging()

    def setUp(self):
        pass

    def tearDown(self):
        pass

    ###################
    # Test constructor
    ###################

    def testConstructor_001(self):
        """
        Test the constructor with device ``/dev/null``, which is writable and exists.
        Use a valid non-ATA SCSI id and defaults for the remaining arguments.  Make sure
        that ``unittest=True``
        """
        writer = CdWriter(device="/dev/null", scsiId="0,0,0", unittest=True)
        self.assertEqual("/dev/null", writer.device)
        self.assertEqual("0,0,0", writer.scsiId)
        self.assertEqual("0,0,0", writer.hardwareId)
        self.assertEqual(None, writer.driveSpeed)
        self.assertEqual(MEDIA_CDRW_74, writer.media.mediaType)
        self.assertEqual(True, writer.isRewritable())
        self.assertEqual(False, writer._noEject)

    def testConstructor_002(self):
        """
        Test the constructor with device ``/dev/null``, which is writable and exists.
        Use a valid ATA SCSI id and defaults for the remaining arguments.  Make sure
        that ``unittest=True``.
        """
        writer = CdWriter(device="/dev/null", scsiId="ATA:0,0,0", unittest=True)
        self.assertEqual("/dev/null", writer.device)
        self.assertEqual("ATA:0,0,0", writer.scsiId)
        self.assertEqual("ATA:0,0,0", writer.hardwareId)
        self.assertEqual(None, writer.driveSpeed)
        self.assertEqual(MEDIA_CDRW_74, writer.media.mediaType)
        self.assertEqual(True, writer.isRewritable())
        self.assertEqual(False, writer._noEject)

    def testConstructor_003(self):
        """
        Test the constructor with device ``/dev/null``, which is writable and exists.
        Use a valid ATAPI SCSI id and defaults for the remaining arguments.  Make sure
        that ``unittest=True``.
        """
        writer = CdWriter(device="/dev/null", scsiId="ATAPI:0,0,0", unittest=True)
        self.assertEqual("/dev/null", writer.device)
        self.assertEqual("ATAPI:0,0,0", writer.scsiId)
        self.assertEqual("ATAPI:0,0,0", writer.hardwareId)
        self.assertEqual(None, writer.driveSpeed)
        self.assertEqual(MEDIA_CDRW_74, writer.media.mediaType)
        self.assertEqual(True, writer.isRewritable())
        self.assertEqual(False, writer._noEject)

    def testConstructor_004(self):
        """
        Test the constructor with device ``/dev/null`` (which is writable and exists).
        Use an invalid SCSI id and defaults for the remaining arguments.  Make sure that
        ``unittest=False``.
        """
        self.assertRaises(ValueError, CdWriter, device="/dev/null", scsiId="blech", unittest=False)

    def testConstructor_005(self):
        """
        Test the constructor with device ``/dev/null`` (which is writable and exists).
        Use an invalid SCSI id and defaults for the remaining arguments.  Make sure that
        ``unittest=True``.
        """
        self.assertRaises(ValueError, CdWriter, device="/dev/null", scsiId="blech", unittest=True)

    def testConstructor_006(self):
        """
        Test the constructor with a non-absolute device path.  Use a valid SCSI
        id and defaults for the remaining arguments.  Make sure that ``unittest=False``.
        """
        self.assertRaises(ValueError, CdWriter, device="dev/null", scsiId="0,0,0", unittest=False)

    def testConstructor_007(self):
        """
        Test the constructor with a non-absolute device path.  Use a valid SCSI
        id and defaults for the remaining arguments.  Make sure that ``unittest=True``.
        """
        self.assertRaises(ValueError, CdWriter, device="dev/null", scsiId="0,0,0", unittest=True)

    def testConstructor_008(self):
        """
        Test the constructor with an absolute device path that does not exist.
        Use a valid SCSI id and defaults for the remaining arguments.  Make sure
        that ``unittest=False``.
        """
        self.assertRaises(ValueError, CdWriter, device="/bogus", scsiId="0,0,0", unittest=False)

    def testConstructor_009(self):
        """
        Test the constructor with an absolute device path that does not exist.
        Use a valid SCSI id and defaults for the remaining arguments.  Make sure
        that ``unittest=True``.
        """
        writer = CdWriter(device="/bogus", scsiId="0,0,0", unittest=True)
        self.assertEqual("/bogus", writer.device)
        self.assertEqual("0,0,0", writer.scsiId)
        self.assertEqual("0,0,0", writer.hardwareId)
        self.assertEqual(None, writer.driveSpeed)
        self.assertEqual(MEDIA_CDRW_74, writer.media.mediaType)
        self.assertEqual(True, writer.isRewritable())
        self.assertEqual(False, writer._noEject)

    def testConstructor_010(self):
        """
        Test the constructor with device ``/dev/null``, which is writable and
        exists.  Use a valid SCSI id and a value of 0 for the drive speed.  Make
        sure that ``unittest=False``.
        """
        self.assertRaises(ValueError, CdWriter, device="/dev/null", scsiId="0,0,0", driveSpeed=0, unittest=False)

    def testConstructor_011(self):
        """
        Test the constructor with device ``/dev/null``, which is writable and
        exists.  Use a valid SCSI id and a value of 0 for the drive speed.  Make
        sure that ``unittest=True``.
        """
        self.assertRaises(ValueError, CdWriter, device="/dev/null", scsiId="0,0,0", driveSpeed=0, unittest=True)

    def testConstructor_012(self):
        """
        Test the constructor with device ``/dev/null``, which is writable and
        exists.  Use a valid SCSI id and a value of 1 for the drive speed.  Make
        sure that ``unittest=True``.
        """
        writer = CdWriter(device="/dev/null", scsiId="0,0,0", driveSpeed=1, unittest=True)
        self.assertEqual("/dev/null", writer.device)
        self.assertEqual("0,0,0", writer.scsiId)
        self.assertEqual("0,0,0", writer.hardwareId)
        self.assertEqual(1, writer.driveSpeed)
        self.assertEqual(MEDIA_CDRW_74, writer.media.mediaType)
        self.assertEqual(True, writer.isRewritable())
        self.assertEqual(False, writer._noEject)

    def testConstructor_013(self):
        """
        Test the constructor with device ``/dev/null``, which is writable and
        exists.  Use a valid SCSI id and a value of 5 for the drive speed.  Make
        sure that ``unittest=True``.
        """
        writer = CdWriter(device="/dev/null", scsiId="0,0,0", driveSpeed=5, unittest=True)
        self.assertEqual("/dev/null", writer.device)
        self.assertEqual("0,0,0", writer.scsiId)
        self.assertEqual("0,0,0", writer.hardwareId)
        self.assertEqual(5, writer.driveSpeed)
        self.assertEqual(MEDIA_CDRW_74, writer.media.mediaType)
        self.assertEqual(True, writer.isRewritable())
        self.assertEqual(False, writer._noEject)

    def testConstructor_014(self):
        """
        Test the constructor with device ``/dev/null``, which is writable and
        exists.  Use a valid SCSI id and an invalid media type.  Make sure that
        ``unittest=False``.
        """
        self.assertRaises(ValueError, CdWriter, device="/dev/null", scsiId="0,0,0", mediaType=42, unittest=False)

    def testConstructor_015(self):
        """
        Test the constructor with device ``/dev/null``, which is writable and
        exists.  Use a valid SCSI id and an invalid media type.  Make sure that
        ``unittest=True``.
        """
        self.assertRaises(ValueError, CdWriter, device="/dev/null", scsiId="0,0,0", mediaType=42, unittest=True)

    def testConstructor_016(self):
        """
        Test the constructor with device ``/dev/null``, which is writable and
        exists.  Use a valid SCSI id and a media type of MEDIA_CDR_74.  Make sure
        that ``unittest=True``.
        """
        writer = CdWriter(device="/dev/null", scsiId="0,0,0", mediaType=MEDIA_CDR_74, unittest=True)
        self.assertEqual("/dev/null", writer.device)
        self.assertEqual("0,0,0", writer.scsiId)
        self.assertEqual("0,0,0", writer.hardwareId)
        self.assertEqual(None, writer.driveSpeed)
        self.assertEqual(MEDIA_CDR_74, writer.media.mediaType)
        self.assertEqual(False, writer.isRewritable())
        self.assertEqual(False, writer._noEject)

    def testConstructor_017(self):
        """
        Test the constructor with device ``/dev/null``, which is writable and
        exists.  Use a valid SCSI id and a media type of MEDIA_CDRW_74.  Make sure
        that ``unittest=True``.
        """
        writer = CdWriter(device="/dev/null", scsiId="0,0,0", mediaType=MEDIA_CDRW_74, unittest=True)
        self.assertEqual("/dev/null", writer.device)
        self.assertEqual("0,0,0", writer.scsiId)
        self.assertEqual("0,0,0", writer.hardwareId)
        self.assertEqual(None, writer.driveSpeed)
        self.assertEqual(MEDIA_CDRW_74, writer.media.mediaType)
        self.assertEqual(True, writer.isRewritable())
        self.assertEqual(False, writer._noEject)

    def testConstructor_018(self):
        """
        Test the constructor with device ``/dev/null``, which is writable and
        exists.  Use a valid SCSI id and a media type of MEDIA_CDR_80.  Make sure
        that ``unittest=True``.
        """
        writer = CdWriter(device="/dev/null", scsiId="0,0,0", mediaType=MEDIA_CDR_80, unittest=True)
        self.assertEqual("/dev/null", writer.device)
        self.assertEqual("0,0,0", writer.scsiId)
        self.assertEqual("0,0,0", writer.hardwareId)
        self.assertEqual(None, writer.driveSpeed)
        self.assertEqual(MEDIA_CDR_80, writer.media.mediaType)
        self.assertEqual(False, writer.isRewritable())
        self.assertEqual(False, writer._noEject)

    def testConstructor_019(self):
        """
        Test the constructor with device ``/dev/null``, which is writable and
        exists.  Use a valid SCSI id and a media type of MEDIA_CDRW_80.  Make sure
        that ``unittest=True``.
        """
        writer = CdWriter(device="/dev/null", scsiId="0,0,0", mediaType=MEDIA_CDRW_80, unittest=True)
        self.assertEqual("/dev/null", writer.device)
        self.assertEqual("0,0,0", writer.scsiId)
        self.assertEqual("0,0,0", writer.hardwareId)
        self.assertEqual(None, writer.driveSpeed)
        self.assertEqual(MEDIA_CDRW_80, writer.media.mediaType)
        self.assertEqual(True, writer.isRewritable())
        self.assertEqual(False, writer._noEject)

    def testConstructor_020(self):
        """
        Test the constructor with device ``/dev/null``, which is writable and
        exists.  Use None for SCSI id and a media type of MEDIA_CDRW_80.  Make
        sure that ``unittest=True``.
        """
        writer = CdWriter(device="/dev/null", scsiId=None, mediaType=MEDIA_CDRW_80, unittest=True)
        self.assertEqual("/dev/null", writer.device)
        self.assertEqual(None, writer.scsiId)
        self.assertEqual("/dev/null", writer.hardwareId)
        self.assertEqual(None, writer.driveSpeed)
        self.assertEqual(MEDIA_CDRW_80, writer.media.mediaType)
        self.assertEqual(True, writer.isRewritable())
        self.assertEqual(False, writer._noEject)

    def testConstructor_021(self):
        """
        Test the constructor with device ``/dev/null``, which is writable and
        exists.  Use None for SCSI id and a media type of MEDIA_CDRW_80.  Make
        sure that ``unittest=True``.  Use ``noEject=True``.
        """
        writer = CdWriter(device="/dev/null", scsiId=None, mediaType=MEDIA_CDRW_80, noEject=True, unittest=True)
        self.assertEqual("/dev/null", writer.device)
        self.assertEqual(None, writer.scsiId)
        self.assertEqual("/dev/null", writer.hardwareId)
        self.assertEqual(None, writer.driveSpeed)
        self.assertEqual(MEDIA_CDRW_80, writer.media.mediaType)
        self.assertEqual(True, writer.isRewritable())
        self.assertEqual(True, writer._noEject)

    ####################################
    # Test the capacity-related methods
    ####################################

    def testCapacity_001(self):
        """
        Test _calculateCapacity for boundaries of None and MEDIA_CDR_74.
        """
        expectedAvailable = MB650 - ILEAD  # 650 MB, minus initial lead-in
        media = MediaDefinition(MEDIA_CDR_74)
        boundaries = None
        capacity = CdWriter._calculateCapacity(media, boundaries)
        self.assertEqual(0, capacity.bytesUsed)
        self.assertEqual(expectedAvailable, capacity.bytesAvailable)
        self.assertEqual(None, capacity.boundaries)

    def testCapacity_002(self):
        """
        Test _calculateCapacity for boundaries of None and MEDIA_CDRW_74.
        """
        expectedAvailable = MB650 - ILEAD  # 650 MB, minus initial lead-in
        media = MediaDefinition(MEDIA_CDRW_74)
        boundaries = None
        capacity = CdWriter._calculateCapacity(media, boundaries)
        self.assertEqual(0, capacity.bytesUsed)
        self.assertEqual(expectedAvailable, capacity.bytesAvailable)
        self.assertEqual(None, capacity.boundaries)

    def testCapacity_003(self):
        """
        Test _calculateCapacity for boundaries of None and MEDIA_CDR_80.
        """
        expectedAvailable = MB700 - ILEAD  # 700 MB, minus initial lead-in
        media = MediaDefinition(MEDIA_CDR_80)
        boundaries = None
        capacity = CdWriter._calculateCapacity(media, boundaries)
        self.assertEqual(0, capacity.bytesUsed)
        self.assertEqual(expectedAvailable, capacity.bytesAvailable)
        self.assertEqual(None, capacity.boundaries)

    def testCapacity_004(self):
        """
        Test _calculateCapacity for boundaries of None and MEDIA_CDRW_80.
        """
        expectedAvailable = MB700 - ILEAD  # 700 MB, minus initial lead-in
        media = MediaDefinition(MEDIA_CDRW_80)
        boundaries = None
        capacity = CdWriter._calculateCapacity(media, boundaries)
        self.assertEqual(0, capacity.bytesUsed)
        self.assertEqual(expectedAvailable, capacity.bytesAvailable)
        self.assertEqual(None, capacity.boundaries)

    def testCapacity_005(self):
        """
        Test _calculateCapacity for boundaries of (0, 1) and MEDIA_CDR_74.
        """
        expectedUsed = 1 * 2048.0  # 1 sector
        expectedAvailable = MB650 - SLEAD - expectedUsed  # 650 MB, minus session lead-in, minus 1 sector
        media = MediaDefinition(MEDIA_CDR_74)
        boundaries = (0, 1)
        capacity = CdWriter._calculateCapacity(media, boundaries)
        self.assertEqual(expectedUsed, capacity.bytesUsed)
        self.assertEqual(expectedAvailable, capacity.bytesAvailable)
        self.assertEqual((0, 1), capacity.boundaries)

    def testCapacity_006(self):
        """
        Test _calculateCapacity for boundaries of (0, 1) and MEDIA_CDRW_74.
        """
        expectedUsed = 1 * 2048.0  # 1 sector
        expectedAvailable = MB650 - SLEAD - expectedUsed  # 650 MB, minus session lead-in, minus 1 sector
        media = MediaDefinition(MEDIA_CDRW_74)
        boundaries = (0, 1)
        capacity = CdWriter._calculateCapacity(media, boundaries)
        self.assertEqual(expectedUsed, capacity.bytesUsed)
        self.assertEqual(expectedAvailable, capacity.bytesAvailable)
        self.assertEqual((0, 1), capacity.boundaries)

    def testCapacity_007(self):
        """
        Test _calculateCapacity for boundaries of (0, 1) and MEDIA_CDR_80.
        """
        expectedUsed = 1 * 2048.0  # 1 sector
        expectedAvailable = MB700 - SLEAD - expectedUsed  # 700 MB, minus session lead-in, minus 1 sector
        media = MediaDefinition(MEDIA_CDR_80)
        boundaries = (0, 1)
        capacity = CdWriter._calculateCapacity(media, boundaries)
        self.assertEqual(expectedUsed, capacity.bytesUsed)
        self.assertEqual(expectedAvailable, capacity.bytesAvailable)  # 700 MB - lead-in - 1 sector
        self.assertEqual((0, 1), capacity.boundaries)

    def testCapacity_008(self):
        """
        Test _calculateCapacity for boundaries of (0, 1) and MEDIA_CDRW_80.
        """
        expectedUsed = 1 * 2048.0  # 1 sector
        expectedAvailable = MB700 - SLEAD - expectedUsed  # 700 MB, minus session lead-in, minus 1 sector
        media = MediaDefinition(MEDIA_CDRW_80)
        boundaries = (0, 1)
        capacity = CdWriter._calculateCapacity(media, boundaries)
        self.assertEqual(expectedUsed, capacity.bytesUsed)
        self.assertEqual(expectedAvailable, capacity.bytesAvailable)
        self.assertEqual((0, 1), capacity.boundaries)

    def testCapacity_009(self):
        """
        Test _calculateCapacity for boundaries of (0, 999) and MEDIA_CDR_74.
        """
        expectedUsed = 999 * 2048.0  # 999 sectors
        expectedAvailable = MB650 - SLEAD - expectedUsed  # 650 MB, minus session lead-in, minus 999 sectors
        media = MediaDefinition(MEDIA_CDR_74)
        boundaries = (0, 999)
        capacity = CdWriter._calculateCapacity(media, boundaries)
        self.assertEqual(expectedUsed, capacity.bytesUsed)
        self.assertEqual(expectedAvailable, capacity.bytesAvailable)
        self.assertEqual((0, 999), capacity.boundaries)

    def testCapacity_010(self):
        """
        Test _calculateCapacity for boundaries of (0, 999) and MEDIA_CDRW_74.
        """
        expectedUsed = 999 * 2048.0  # 999 sectors
        expectedAvailable = MB650 - SLEAD - expectedUsed  # 650 MB, minus session lead-in, minus 999 sectors
        media = MediaDefinition(MEDIA_CDRW_74)
        boundaries = (0, 999)
        capacity = CdWriter._calculateCapacity(media, boundaries)
        self.assertEqual(expectedUsed, capacity.bytesUsed)
        self.assertEqual(expectedAvailable, capacity.bytesAvailable)
        self.assertEqual((0, 999), capacity.boundaries)

    def testCapacity_011(self):
        """
        Test _calculateCapacity for boundaries of (0, 999) and MEDIA_CDR_80.
        """
        expectedUsed = 999 * 2048.0  # 999 sectors
        expectedAvailable = MB700 - SLEAD - expectedUsed  # 700 MB, minus session lead-in, minus 999 sectors
        media = MediaDefinition(MEDIA_CDR_80)
        boundaries = (0, 999)
        capacity = CdWriter._calculateCapacity(media, boundaries)
        self.assertEqual(expectedUsed, capacity.bytesUsed)
        self.assertEqual(expectedAvailable, capacity.bytesAvailable)
        self.assertEqual((0, 999), capacity.boundaries)

    def testCapacity_012(self):
        """
        Test _calculateCapacity for boundaries of (0, 999) and MEDIA_CDRW_80.
        """
        expectedUsed = 999 * 2048.0  # 999 sectors
        expectedAvailable = MB700 - SLEAD - expectedUsed  # 700 MB, minus session lead-in, minus 999 sectors
        media = MediaDefinition(MEDIA_CDRW_80)
        boundaries = (0, 999)
        capacity = CdWriter._calculateCapacity(media, boundaries)
        self.assertEqual(expectedUsed, capacity.bytesUsed)
        self.assertEqual(expectedAvailable, capacity.bytesAvailable)
        self.assertEqual((0, 999), capacity.boundaries)

    def testCapacity_013(self):
        """
        Test _calculateCapacity for boundaries of (500, 1000) and MEDIA_CDR_74.
        """
        expectedUsed = 1000 * 2048.0  # 1000 sectors
        expectedAvailable = MB650 - SLEAD - expectedUsed  # 650 MB, minus session lead-in, minus 1000 sectors
        media = MediaDefinition(MEDIA_CDR_74)
        boundaries = (500, 1000)
        capacity = CdWriter._calculateCapacity(media, boundaries)
        self.assertEqual(expectedUsed, capacity.bytesUsed)
        self.assertEqual(expectedAvailable, capacity.bytesAvailable)
        self.assertEqual((500, 1000), capacity.boundaries)

    def testCapacity_014(self):
        """
        Test _calculateCapacity for boundaries of (500, 1000) and MEDIA_CDRW_74.
        """
        expectedUsed = 1000 * 2048.0  # 1000 sectors
        expectedAvailable = MB650 - SLEAD - expectedUsed  # 650 MB, minus session lead-in, minus 1000 sectors
        media = MediaDefinition(MEDIA_CDRW_74)
        boundaries = (500, 1000)
        capacity = CdWriter._calculateCapacity(media, boundaries)
        self.assertEqual(expectedUsed, capacity.bytesUsed)
        self.assertEqual(expectedAvailable, capacity.bytesAvailable)
        self.assertEqual((500, 1000), capacity.boundaries)

    def testCapacity_015(self):
        """
        Test _calculateCapacity for boundaries of (500, 1000) and MEDIA_CDR_80.
        """
        expectedUsed = 1000 * 2048.0  # 1000 sectors
        expectedAvailable = MB700 - SLEAD - expectedUsed  # 700 MB, minus session lead-in, minus 1000 sectors
        media = MediaDefinition(MEDIA_CDR_80)
        boundaries = (500, 1000)
        capacity = CdWriter._calculateCapacity(media, boundaries)
        self.assertEqual(expectedUsed, capacity.bytesUsed)
        self.assertEqual(expectedAvailable, capacity.bytesAvailable)
        self.assertEqual((500, 1000), capacity.boundaries)

    def testCapacity_016(self):
        """
        Test _calculateCapacity for boundaries of (500, 1000) and MEDIA_CDRW_80.
        """
        expectedUsed = 1000 * 2048.0  # 1000 sectors
        expectedAvailable = MB700 - SLEAD - expectedUsed  # 700 MB, minus session lead-in, minus 1000 sectors
        media = MediaDefinition(MEDIA_CDRW_80)
        boundaries = (500, 1000)
        capacity = CdWriter._calculateCapacity(media, boundaries)
        self.assertEqual(expectedUsed, capacity.bytesUsed)
        self.assertEqual(expectedAvailable, capacity.bytesAvailable)  # 650 MB minus lead-in
        self.assertEqual((500, 1000), capacity.boundaries)

    def testCapacity_017(self):
        """
        Test _getBoundaries when self.deviceSupportsMulti is False; entireDisc=False, useMulti=True.
        """
        writer = CdWriter(device="/dev/cdrw", scsiId="0,0,0", unittest=True)
        writer._deviceSupportsMulti = False
        boundaries = writer._getBoundaries(entireDisc=False, useMulti=True)
        self.assertEqual(None, boundaries)

    def testCapacity_018(self):
        """
        Test _getBoundaries when self.deviceSupportsMulti is False; entireDisc=True, useMulti=True.
        """
        writer = CdWriter(device="/dev/cdrw", scsiId="0,0,0", unittest=True)
        writer._deviceSupportsMulti = False
        boundaries = writer._getBoundaries(entireDisc=True, useMulti=True)
        self.assertEqual(None, boundaries)

    def testCapacity_019(self):
        """
        Test _getBoundaries when self.deviceSupportsMulti is False; entireDisc=True, useMulti=False.
        """
        writer = CdWriter(device="/dev/cdrw", scsiId="0,0,0", unittest=True)
        writer._deviceSupportsMulti = False
        boundaries = writer._getBoundaries(entireDisc=False, useMulti=False)
        self.assertEqual(None, boundaries)

    def testCapacity_020(self):
        """
        Test _getBoundaries when self.deviceSupportsMulti is False; entireDisc=False, useMulti=False.
        """
        writer = CdWriter(device="/dev/cdrw", scsiId="0,0,0", unittest=True)
        writer._deviceSupportsMulti = False
        boundaries = writer._getBoundaries(entireDisc=False, useMulti=False)
        self.assertEqual(None, boundaries)

    def testCapacity_021(self):
        """
        Test _getBoundaries when self.deviceSupportsMulti is True; entireDisc=True, useMulti=True.
        """
        writer = CdWriter(device="/dev/cdrw", scsiId="0,0,0", unittest=True)
        writer._deviceSupportsMulti = True
        boundaries = writer._getBoundaries(entireDisc=True, useMulti=True)
        self.assertEqual(None, boundaries)

    def testCapacity_022(self):
        """
        Test _getBoundaries when self.deviceSupportsMulti is True; entireDisc=True, useMulti=False.
        """
        writer = CdWriter(device="/dev/cdrw", scsiId="0,0,0", unittest=True)
        writer._deviceSupportsMulti = True
        boundaries = writer._getBoundaries(entireDisc=True, useMulti=False)
        self.assertEqual(None, boundaries)

    def testCapacity_023(self):
        """
        Test _calculateCapacity for boundaries of (321342, 330042) and MEDIA_CDRW_74.
        This was a bug fixed for v2.1.2.
        """
        expectedUsed = 330042 * 2048.0  # 330042 sectors
        expectedAvailable = 0  # nothing should be available
        media = MediaDefinition(MEDIA_CDRW_74)
        boundaries = (321342, 330042)
        capacity = CdWriter._calculateCapacity(media, boundaries)
        self.assertEqual(expectedUsed, capacity.bytesUsed)
        self.assertEqual(expectedAvailable, capacity.bytesAvailable)
        self.assertEqual((321342, 330042), capacity.boundaries)

    def testCapacity_024(self):
        """
        Test _calculateCapacity for boundaries of (0, 330042) and MEDIA_CDRW_74.
        This was a bug fixed for v2.1.3.
        """
        expectedUsed = 330042 * 2048.0  # 330042 sectors
        expectedAvailable = 0  # nothing should be available
        media = MediaDefinition(MEDIA_CDRW_74)
        boundaries = (0, 330042)
        capacity = CdWriter._calculateCapacity(media, boundaries)
        self.assertEqual(expectedUsed, capacity.bytesUsed)
        self.assertEqual(expectedAvailable, capacity.bytesAvailable)
        self.assertEqual((0, 330042), capacity.boundaries)

    #########################################
    # Test methods that build argument lists
    #########################################

    def testBuildArgs_001(self):
        """
        Test _buildOpenTrayArgs().
        """
        args = CdWriter._buildOpenTrayArgs(device="/dev/stuff")
        self.assertEqual(["/dev/stuff"], args)

    def testBuildArgs_002(self):
        """
        Test _buildCloseTrayArgs().
        """
        args = CdWriter._buildCloseTrayArgs(device="/dev/stuff")
        self.assertEqual(["-t", "/dev/stuff"], args)

    def testBuildArgs_003(self):
        """
        Test _buildPropertiesArgs().
        """
        args = CdWriter._buildPropertiesArgs(hardwareId="0,0,0")
        self.assertEqual(["-prcap", "dev=0,0,0"], args)

    def testBuildArgs_004(self):
        """
        Test _buildBoundariesArgs().
        """
        args = CdWriter._buildBoundariesArgs(hardwareId="ATA:0,0,0")
        self.assertEqual(["-msinfo", "dev=ATA:0,0,0"], args)

    def testBuildArgs_005(self):
        """
        Test _buildBoundariesArgs().
        """
        args = CdWriter._buildBoundariesArgs(hardwareId="ATAPI:0,0,0")
        self.assertEqual(["-msinfo", "dev=ATAPI:0,0,0"], args)

    def testBuildArgs_006(self):
        """
        Test _buildBlankArgs(), default drive speed.
        """
        args = CdWriter._buildBlankArgs(hardwareId="ATA:0,0,0")
        self.assertEqual(["-v", "blank=fast", "dev=ATA:0,0,0"], args)

    def testBuildArgs_007(self):
        """
        Test _buildBlankArgs(), default drive speed.
        """
        args = CdWriter._buildBlankArgs(hardwareId="ATAPI:0,0,0")
        self.assertEqual(["-v", "blank=fast", "dev=ATAPI:0,0,0"], args)

    def testBuildArgs_008(self):
        """
        Test _buildBlankArgs(), with None for drive speed.
        """
        args = CdWriter._buildBlankArgs(hardwareId="0,0,0", driveSpeed=None)
        self.assertEqual(["-v", "blank=fast", "dev=0,0,0"], args)

    def testBuildArgs_009(self):
        """
        Test _buildBlankArgs(), with 1 for drive speed.
        """
        args = CdWriter._buildBlankArgs(hardwareId="0,0,0", driveSpeed=1)
        self.assertEqual(["-v", "blank=fast", "speed=1", "dev=0,0,0"], args)

    def testBuildArgs_010(self):
        """
        Test _buildBlankArgs(), with 5 for drive speed.
        """
        args = CdWriter._buildBlankArgs(hardwareId="ATA:1,2,3", driveSpeed=5)
        self.assertEqual(["-v", "blank=fast", "speed=5", "dev=ATA:1,2,3"], args)

    def testBuildArgs_011(self):
        """
        Test _buildBlankArgs(), with 5 for drive speed.
        """
        args = CdWriter._buildBlankArgs(hardwareId="ATAPI:1,2,3", driveSpeed=5)
        self.assertEqual(["-v", "blank=fast", "speed=5", "dev=ATAPI:1,2,3"], args)

    def testBuildArgs_012(self):
        """
        Test _buildWriteArgs(), default drive speed and writeMulti.
        """
        args = CdWriter._buildWriteArgs(hardwareId="0,0,0", imagePath="/whatever")
        self.assertEqual(["-v", "dev=0,0,0", "-multi", "-data", "/whatever"], args)

    def testBuildArgs_013(self):
        """
        Test _buildWriteArgs(), None for drive speed, True for writeMulti.
        """
        args = CdWriter._buildWriteArgs(hardwareId="0,0,0", imagePath="/whatever", driveSpeed=None, writeMulti=True)
        self.assertEqual(["-v", "dev=0,0,0", "-multi", "-data", "/whatever"], args)

    def testBuildArgs_014(self):
        """
        Test _buildWriteArgs(), None for drive speed, False for writeMulti.
        """
        args = CdWriter._buildWriteArgs(hardwareId="0,0,0", imagePath="/whatever", driveSpeed=None, writeMulti=False)
        self.assertEqual(["-v", "dev=0,0,0", "-data", "/whatever"], args)

    def testBuildArgs_015(self):
        """
        Test _buildWriteArgs(), 1 for drive speed, True for writeMulti.
        """
        args = CdWriter._buildWriteArgs(hardwareId="0,0,0", imagePath="/whatever", driveSpeed=1, writeMulti=True)
        self.assertEqual(["-v", "speed=1", "dev=0,0,0", "-multi", "-data", "/whatever"], args)

    def testBuildArgs_016(self):
        """
        Test _buildWriteArgs(), 5 for drive speed, True for writeMulti.
        """
        args = CdWriter._buildWriteArgs(hardwareId="0,1,2", imagePath="/whatever", driveSpeed=5, writeMulti=True)
        self.assertEqual(["-v", "speed=5", "dev=0,1,2", "-multi", "-data", "/whatever"], args)

    def testBuildArgs_017(self):
        """
        Test _buildWriteArgs(), 1 for drive speed, False for writeMulti.
        """
        args = CdWriter._buildWriteArgs(hardwareId="0,0,0", imagePath="/dvl/stuff/whatever/more", driveSpeed=1, writeMulti=False)
        self.assertEqual(["-v", "speed=1", "dev=0,0,0", "-data", "/dvl/stuff/whatever/more"], args)

    def testBuildArgs_018(self):
        """
        Test _buildWriteArgs(), 5 for drive speed, False for writeMulti.
        """
        args = CdWriter._buildWriteArgs(hardwareId="ATA:1,2,3", imagePath="/whatever", driveSpeed=5, writeMulti=False)
        self.assertEqual(["-v", "speed=5", "dev=ATA:1,2,3", "-data", "/whatever"], args)

    def testBuildArgs_019(self):
        """
        Test _buildWriteArgs(), 5 for drive speed, False for writeMulti.
        """
        args = CdWriter._buildWriteArgs(hardwareId="ATAPI:1,2,3", imagePath="/whatever", driveSpeed=5, writeMulti=False)
        self.assertEqual(["-v", "speed=5", "dev=ATAPI:1,2,3", "-data", "/whatever"], args)

    ##########################################
    # Test methods that parse cdrecord output
    ##########################################

    def testParseOutput_001(self):
        """
        Test _parseBoundariesOutput() for valid data, taken from a real example.
        """
        output = [
            "268582,302230\n",
        ]
        boundaries = CdWriter._parseBoundariesOutput(output)
        self.assertEqual((268582, 302230), boundaries)

    def testParseOutput_002(self):
        """
        Test _parseBoundariesOutput() for valid data, taken from a real example,
        lots of extra whitespace around the values.
        """
        output = [
            "   268582 ,  302230    \n",
        ]
        boundaries = CdWriter._parseBoundariesOutput(output)
        self.assertEqual((268582, 302230), boundaries)

    def testParseOutput_003(self):
        """
        Test _parseBoundariesOutput() for valid data, taken from a real example,
        lots of extra garbage after the first line.
        """
        output = [
            "268582,302230\n",
            "more\n",
            "bogus\n",
            "crap\n",
            "here\n",
            "to\n",
            "confuse\n",
            "things\n",
        ]
        boundaries = CdWriter._parseBoundariesOutput(output)
        self.assertEqual((268582, 302230), boundaries)

    def testParseOutput_004(self):
        """
        Test _parseBoundariesOutput() for valid data, taken from a real example,
        lots of extra garbage before the first line.
        """
        output = [
            "more\n",
            "bogus\n",
            "crap\n",
            "here\n",
            "to\n",
            "confuse\n",
            "things\n",
            "268582,302230\n",
        ]
        self.assertRaises(IOError, CdWriter._parseBoundariesOutput, output)

    def testParseOutput_005(self):
        """
        Test _parseBoundariesOutput() for valid data, taken from a real example,
        with first value converted to negative.
        """
        output = [
            "-268582,302230\n",
        ]
        self.assertRaises(IOError, CdWriter._parseBoundariesOutput, output)

    def testParseOutput_006(self):
        """
        Test _parseBoundariesOutput() for valid data, taken from a real example,
        with second value converted to negative.
        """
        output = [
            "268582,-302230\n",
        ]
        self.assertRaises(IOError, CdWriter._parseBoundariesOutput, output)

    def testParseOutput_007(self):
        """
        Test _parseBoundariesOutput() for valid data, taken from a real example,
        with first value converted to zero.
        """
        output = [
            "0,302230\n",
        ]
        boundaries = CdWriter._parseBoundariesOutput(output)
        self.assertEqual((0, 302230), boundaries)

    def testParseOutput_008(self):
        """
        Test _parseBoundariesOutput() for valid data, taken from a real example,
        with second value converted to zero.
        """
        output = [
            "268582,0\n",
        ]
        boundaries = CdWriter._parseBoundariesOutput(output)
        self.assertEqual((268582, 0), boundaries)

    def testParseOutput_009(self):
        """
        Test _parseBoundariesOutput() for valid data, taken from a real example,
        with first value converted to negative and second value converted to zero.
        """
        output = [
            "-268582,0\n",
        ]
        self.assertRaises(IOError, CdWriter._parseBoundariesOutput, output)

    def testParseOutput_010(self):
        """
        Test _parseBoundariesOutput() for valid data, taken from a real example,
        with first value converted to zero and second value converted to negative.
        """
        output = [
            "0,-302230\n",
        ]
        self.assertRaises(IOError, CdWriter._parseBoundariesOutput, output)

    def testParseOutput_011(self):
        """
        Test _parsePropertiesOutput() for valid data, taken from a real example,
        including stderr and stdout mixed together.
        """
        output = [
            "scsidev: '0,0,0'\n",
            "scsibus: 0 target: 0 lun: 0\n",
            "Linux sg driver version: 3.1.22\n",
            "Cdrecord 1.10 (i686-pc-linux-gnu) Copyright (C) 1995-2001 J\xf6rg Schilling\n",
            "Using libscg version 'schily-0.5'\n",
            "Device type    : Removable CD-ROM\n",
            "Version        : 0\n",
            "Response Format: 1\n",
            "Vendor_info    : 'SONY    '\n",
            "Identifikation : 'CD-RW  CRX140E  '\n",
            "Revision       : '1.0n'\n",
            "Device seems to be: Generic mmc CD-RW.\n",
            "\n",
            "Drive capabilities, per page 2A:\n",
            "\n",
            "  Does read CD-R media\n",
            "  Does write CD-R media\n",
            "  Does read CD-RW media\n",
            "  Does write CD-RW media\n",
            "  Does not read DVD-ROM media\n",
            "  Does not read DVD-R media\n",
            "  Does not write DVD-R media\n",
            "  Does not read DVD-RAM media\n",
            "  Does not write DVD-RAM media\n",
            "  Does support test writing\n",
            "\n",
            "  Does read Mode 2 Form 1 blocks\n",
            "  Does read Mode 2 Form 2 blocks\n",
            "  Does read digital audio blocks\n",
            "  Does restart non-streamed digital audio reads accurately\n",
            "  Does not support BURN-Proof (Sanyo)\n",
            "  Does read multi-session CDs\n",
            "  Does read fixed-packet CD media using Method 2\n",
            "  Does not read CD bar code\n",
            "  Does not read R-W subcode information\n",
            "  Does read raw P-W subcode data from lead in\n",
            "  Does return CD media catalog number\n",
            "  Does return CD ISRC information\n",
            "  Does not support C2 error pointers\n",
            "  Does not deliver composite A/V data\n",
            "\n",
            "  Does play audio CDs\n",
            "  Number of volume control levels: 256\n",
            "  Does support individual volume control setting for each channel\n",
            "  Does support independent mute setting for each channel\n",
            "  Does not support digital output on port 1\n",
            "  Does not support digital output on port 2\n",
            "\n",
            "  Loading mechanism type: tray\n",
            "  Does support ejection of CD via START/STOP command\n",
            "  Does not lock media on power up via prevent jumper\n",
            "  Does allow media to be locked in the drive via PREVENT/ALLOW command\n",
            "  Is not currently in a media-locked state\n",
            "  Does not support changing side of disk\n",
            "  Does not have load-empty-slot-in-changer feature\n",
            "  Does not support Individual Disk Present feature\n",
            "\n",
            "  Maximum read  speed in kB/s: 5645\n",
            "  Current read  speed in kB/s: 3528\n",
            "  Maximum write speed in kB/s: 1411\n",
            "  Current write speed in kB/s: 706\n",
            "  Buffer size in KB: 4096\n",
        ]
        (
            deviceType,
            deviceVendor,
            deviceId,
            deviceBufferSize,
            deviceSupportsMulti,
            deviceHasTray,
            deviceCanEject,
        ) = CdWriter._parsePropertiesOutput(output)
        self.assertEqual("Removable CD-ROM", deviceType)
        self.assertEqual("SONY", deviceVendor)
        self.assertEqual("CD-RW  CRX140E", deviceId)
        self.assertEqual(4096.0 * 1024.0, deviceBufferSize)
        self.assertEqual(True, deviceSupportsMulti)
        self.assertEqual(True, deviceHasTray)
        self.assertEqual(True, deviceCanEject)

    def testParseOutput_012(self):
        """
        Test _parsePropertiesOutput() for valid data, taken from a real example,
        including only stdout.
        """
        output = [
            "Cdrecord 1.10 (i686-pc-linux-gnu) Copyright (C) 1995-2001 J\xf6rg Schilling\n",
            "Using libscg version 'schily-0.5'\n",
            "Device type    : Removable CD-ROM\n",
            "Version        : 0\n",
            "Response Format: 1\n",
            "Vendor_info    : 'SONY    '\n",
            "Identifikation : 'CD-RW  CRX140E  '\n",
            "Revision       : '1.0n'\n",
            "Device seems to be: Generic mmc CD-RW.\n",
            "\n",
            "Drive capabilities, per page 2A:\n",
            "\n",
            "  Does read CD-R media\n",
            "  Does write CD-R media\n",
            "  Does read CD-RW media\n",
            "  Does write CD-RW media\n",
            "  Does not read DVD-ROM media\n",
            "  Does not read DVD-R media\n",
            "  Does not write DVD-R media\n",
            "  Does not read DVD-RAM media\n",
            "  Does not write DVD-RAM media\n",
            "  Does support test writing\n",
            "\n",
            "  Does read Mode 2 Form 1 blocks\n",
            "  Does read Mode 2 Form 2 blocks\n",
            "  Does read digital audio blocks\n",
            "  Does restart non-streamed digital audio reads accurately\n",
            "  Does not support BURN-Proof (Sanyo)\n",
            "  Does read multi-session CDs\n",
            "  Does read fixed-packet CD media using Method 2\n",
            "  Does not read CD bar code\n",
            "  Does not read R-W subcode information\n",
            "  Does read raw P-W subcode data from lead in\n",
            "  Does return CD media catalog number\n",
            "  Does return CD ISRC information\n",
            "  Does not support C2 error pointers\n",
            "  Does not deliver composite A/V data\n",
            "\n",
            "  Does play audio CDs\n",
            "  Number of volume control levels: 256\n",
            "  Does support individual volume control setting for each channel\n",
            "  Does support independent mute setting for each channel\n",
            "  Does not support digital output on port 1\n",
            "  Does not support digital output on port 2\n",
            "\n",
            "  Loading mechanism type: tray\n",
            "  Does support ejection of CD via START/STOP command\n",
            "  Does not lock media on power up via prevent jumper\n",
            "  Does allow media to be locked in the drive via PREVENT/ALLOW command\n",
            "  Is not currently in a media-locked state\n",
            "  Does not support changing side of disk\n",
            "  Does not have load-empty-slot-in-changer feature\n",
            "  Does not support Individual Disk Present feature\n",
            "\n",
            "  Maximum read  speed in kB/s: 5645\n",
            "  Current read  speed in kB/s: 3528\n",
            "  Maximum write speed in kB/s: 1411\n",
            "  Current write speed in kB/s: 706\n",
            "  Buffer size in KB: 4096\n",
        ]
        (
            deviceType,
            deviceVendor,
            deviceId,
            deviceBufferSize,
            deviceSupportsMulti,
            deviceHasTray,
            deviceCanEject,
        ) = CdWriter._parsePropertiesOutput(output)
        self.assertEqual("Removable CD-ROM", deviceType)
        self.assertEqual("SONY", deviceVendor)
        self.assertEqual("CD-RW  CRX140E", deviceId)
        self.assertEqual(4096.0 * 1024.0, deviceBufferSize)
        self.assertEqual(True, deviceSupportsMulti)
        self.assertEqual(True, deviceHasTray)
        self.assertEqual(True, deviceCanEject)

    def testParseOutput_013(self):
        """
        Test _parsePropertiesOutput() for valid data, taken from a real example,
        including stderr and stdout mixed together, device type removed.
        """
        output = [
            "scsidev: '0,0,0'\n",
            "scsibus: 0 target: 0 lun: 0\n",
            "Linux sg driver version: 3.1.22\n",
            "Cdrecord 1.10 (i686-pc-linux-gnu) Copyright (C) 1995-2001 J\xf6rg Schilling\n",
            "Using libscg version 'schily-0.5'\n",
            "Version        : 0\n",
            "Response Format: 1\n",
            "Vendor_info    : 'SONY    '\n",
            "Identifikation : 'CD-RW  CRX140E  '\n",
            "Revision       : '1.0n'\n",
            "Device seems to be: Generic mmc CD-RW.\n",
            "\n",
            "Drive capabilities, per page 2A:\n",
            "\n",
            "  Does read CD-R media\n",
            "  Does write CD-R media\n",
            "  Does read CD-RW media\n",
            "  Does write CD-RW media\n",
            "  Does not read DVD-ROM media\n",
            "  Does not read DVD-R media\n",
            "  Does not write DVD-R media\n",
            "  Does not read DVD-RAM media\n",
            "  Does not write DVD-RAM media\n",
            "  Does support test writing\n",
            "\n",
            "  Does read Mode 2 Form 1 blocks\n",
            "  Does read Mode 2 Form 2 blocks\n",
            "  Does read digital audio blocks\n",
            "  Does restart non-streamed digital audio reads accurately\n",
            "  Does not support BURN-Proof (Sanyo)\n",
            "  Does read multi-session CDs\n",
            "  Does read fixed-packet CD media using Method 2\n",
            "  Does not read CD bar code\n",
            "  Does not read R-W subcode information\n",
            "  Does read raw P-W subcode data from lead in\n",
            "  Does return CD media catalog number\n",
            "  Does return CD ISRC information\n",
            "  Does not support C2 error pointers\n",
            "  Does not deliver composite A/V data\n",
            "\n",
            "  Does play audio CDs\n",
            "  Number of volume control levels: 256\n",
            "  Does support individual volume control setting for each channel\n",
            "  Does support independent mute setting for each channel\n",
            "  Does not support digital output on port 1\n",
            "  Does not support digital output on port 2\n",
            "\n",
            "  Loading mechanism type: tray\n",
            "  Does support ejection of CD via START/STOP command\n",
            "  Does not lock media on power up via prevent jumper\n",
            "  Does allow media to be locked in the drive via PREVENT/ALLOW command\n",
            "  Is not currently in a media-locked state\n",
            "  Does not support changing side of disk\n",
            "  Does not have load-empty-slot-in-changer feature\n",
            "  Does not support Individual Disk Present feature\n",
            "\n",
            "  Maximum read  speed in kB/s: 5645\n",
            "  Current read  speed in kB/s: 3528\n",
            "  Maximum write speed in kB/s: 1411\n",
            "  Current write speed in kB/s: 706\n",
            "  Buffer size in KB: 4096\n",
        ]
        (
            deviceType,
            deviceVendor,
            deviceId,
            deviceBufferSize,
            deviceSupportsMulti,
            deviceHasTray,
            deviceCanEject,
        ) = CdWriter._parsePropertiesOutput(output)
        self.assertEqual(None, deviceType)
        self.assertEqual("SONY", deviceVendor)
        self.assertEqual("CD-RW  CRX140E", deviceId)
        self.assertEqual(4096.0 * 1024.0, deviceBufferSize)
        self.assertEqual(True, deviceSupportsMulti)
        self.assertEqual(True, deviceHasTray)
        self.assertEqual(True, deviceCanEject)

    def testParseOutput_014(self):
        """
        Test _parsePropertiesOutput() for valid data, taken from a real example,
        including stderr and stdout mixed together, device vendor removed.
        """
        output = [
            "scsidev: '0,0,0'\n",
            "scsibus: 0 target: 0 lun: 0\n",
            "Linux sg driver version: 3.1.22\n",
            "Cdrecord 1.10 (i686-pc-linux-gnu) Copyright (C) 1995-2001 J\xf6rg Schilling\n",
            "Using libscg version 'schily-0.5'\n",
            "Device type    : Removable CD-ROM\n",
            "Version        : 0\n",
            "Response Format: 1\n",
            "Identifikation : 'CD-RW  CRX140E  '\n",
            "Revision       : '1.0n'\n",
            "Device seems to be: Generic mmc CD-RW.\n",
            "\n",
            "Drive capabilities, per page 2A:\n",
            "\n",
            "  Does read CD-R media\n",
            "  Does write CD-R media\n",
            "  Does read CD-RW media\n",
            "  Does write CD-RW media\n",
            "  Does not read DVD-ROM media\n",
            "  Does not read DVD-R media\n",
            "  Does not write DVD-R media\n",
            "  Does not read DVD-RAM media\n",
            "  Does not write DVD-RAM media\n",
            "  Does support test writing\n",
            "\n",
            "  Does read Mode 2 Form 1 blocks\n",
            "  Does read Mode 2 Form 2 blocks\n",
            "  Does read digital audio blocks\n",
            "  Does restart non-streamed digital audio reads accurately\n",
            "  Does not support BURN-Proof (Sanyo)\n",
            "  Does read multi-session CDs\n",
            "  Does read fixed-packet CD media using Method 2\n",
            "  Does not read CD bar code\n",
            "  Does not read R-W subcode information\n",
            "  Does read raw P-W subcode data from lead in\n",
            "  Does return CD media catalog number\n",
            "  Does return CD ISRC information\n",
            "  Does not support C2 error pointers\n",
            "  Does not deliver composite A/V data\n",
            "\n",
            "  Does play audio CDs\n",
            "  Number of volume control levels: 256\n",
            "  Does support individual volume control setting for each channel\n",
            "  Does support independent mute setting for each channel\n",
            "  Does not support digital output on port 1\n",
            "  Does not support digital output on port 2\n",
            "\n",
            "  Loading mechanism type: tray\n",
            "  Does support ejection of CD via START/STOP command\n",
            "  Does not lock media on power up via prevent jumper\n",
            "  Does allow media to be locked in the drive via PREVENT/ALLOW command\n",
            "  Is not currently in a media-locked state\n",
            "  Does not support changing side of disk\n",
            "  Does not have load-empty-slot-in-changer feature\n",
            "  Does not support Individual Disk Present feature\n",
            "\n",
            "  Maximum read  speed in kB/s: 5645\n",
            "  Current read  speed in kB/s: 3528\n",
            "  Maximum write speed in kB/s: 1411\n",
            "  Current write speed in kB/s: 706\n",
            "  Buffer size in KB: 4096\n",
        ]
        (
            deviceType,
            deviceVendor,
            deviceId,
            deviceBufferSize,
            deviceSupportsMulti,
            deviceHasTray,
            deviceCanEject,
        ) = CdWriter._parsePropertiesOutput(output)
        self.assertEqual("Removable CD-ROM", deviceType)
        self.assertEqual(None, deviceVendor)
        self.assertEqual("CD-RW  CRX140E", deviceId)
        self.assertEqual(4096.0 * 1024.0, deviceBufferSize)
        self.assertEqual(True, deviceSupportsMulti)
        self.assertEqual(True, deviceHasTray)
        self.assertEqual(True, deviceCanEject)

    def testParseOutput_015(self):
        """
        Test _parsePropertiesOutput() for valid data, taken from a real example,
        including stderr and stdout mixed together, device id removed.
        """
        output = [
            "scsidev: '0,0,0'\n",
            "scsibus: 0 target: 0 lun: 0\n",
            "Linux sg driver version: 3.1.22\n",
            "Cdrecord 1.10 (i686-pc-linux-gnu) Copyright (C) 1995-2001 J\xf6rg Schilling\n",
            "Using libscg version 'schily-0.5'\n",
            "Device type    : Removable CD-ROM\n",
            "Version        : 0\n",
            "Response Format: 1\n",
            "Vendor_info    : 'SONY    '\n",
            "Revision       : '1.0n'\n",
            "Device seems to be: Generic mmc CD-RW.\n",
            "\n",
            "Drive capabilities, per page 2A:\n",
            "\n",
            "  Does read CD-R media\n",
            "  Does write CD-R media\n",
            "  Does read CD-RW media\n",
            "  Does write CD-RW media\n",
            "  Does not read DVD-ROM media\n",
            "  Does not read DVD-R media\n",
            "  Does not write DVD-R media\n",
            "  Does not read DVD-RAM media\n",
            "  Does not write DVD-RAM media\n",
            "  Does support test writing\n",
            "\n",
            "  Does read Mode 2 Form 1 blocks\n",
            "  Does read Mode 2 Form 2 blocks\n",
            "  Does read digital audio blocks\n",
            "  Does restart non-streamed digital audio reads accurately\n",
            "  Does not support BURN-Proof (Sanyo)\n",
            "  Does read multi-session CDs\n",
            "  Does read fixed-packet CD media using Method 2\n",
            "  Does not read CD bar code\n",
            "  Does not read R-W subcode information\n",
            "  Does read raw P-W subcode data from lead in\n",
            "  Does return CD media catalog number\n",
            "  Does return CD ISRC information\n",
            "  Does not support C2 error pointers\n",
            "  Does not deliver composite A/V data\n",
            "\n",
            "  Does play audio CDs\n",
            "  Number of volume control levels: 256\n",
            "  Does support individual volume control setting for each channel\n",
            "  Does support independent mute setting for each channel\n",
            "  Does not support digital output on port 1\n",
            "  Does not support digital output on port 2\n",
            "\n",
            "  Loading mechanism type: tray\n",
            "  Does support ejection of CD via START/STOP command\n",
            "  Does not lock media on power up via prevent jumper\n",
            "  Does allow media to be locked in the drive via PREVENT/ALLOW command\n",
            "  Is not currently in a media-locked state\n",
            "  Does not support changing side of disk\n",
            "  Does not have load-empty-slot-in-changer feature\n",
            "  Does not support Individual Disk Present feature\n",
            "\n",
            "  Maximum read  speed in kB/s: 5645\n",
            "  Current read  speed in kB/s: 3528\n",
            "  Maximum write speed in kB/s: 1411\n",
            "  Current write speed in kB/s: 706\n",
            "  Buffer size in KB: 4096\n",
        ]
        (
            deviceType,
            deviceVendor,
            deviceId,
            deviceBufferSize,
            deviceSupportsMulti,
            deviceHasTray,
            deviceCanEject,
        ) = CdWriter._parsePropertiesOutput(output)
        self.assertEqual("Removable CD-ROM", deviceType)
        self.assertEqual("SONY", deviceVendor)
        self.assertEqual(None, deviceId)
        self.assertEqual(4096.0 * 1024.0, deviceBufferSize)
        self.assertEqual(True, deviceSupportsMulti)
        self.assertEqual(True, deviceHasTray)
        self.assertEqual(True, deviceCanEject)

    def testParseOutput_016(self):
        """
        Test _parsePropertiesOutput() for valid data, taken from a real example,
        including stderr and stdout mixed together, buffer size removed.
        """
        output = [
            "scsidev: '0,0,0'\n",
            "scsibus: 0 target: 0 lun: 0\n",
            "Linux sg driver version: 3.1.22\n",
            "Cdrecord 1.10 (i686-pc-linux-gnu) Copyright (C) 1995-2001 J\xf6rg Schilling\n",
            "Using libscg version 'schily-0.5'\n",
            "Device type    : Removable CD-ROM\n",
            "Version        : 0\n",
            "Response Format: 1\n",
            "Vendor_info    : 'SONY    '\n",
            "Identifikation : 'CD-RW  CRX140E  '\n",
            "Revision       : '1.0n'\n",
            "Device seems to be: Generic mmc CD-RW.\n",
            "\n",
            "Drive capabilities, per page 2A:\n",
            "\n",
            "  Does read CD-R media\n",
            "  Does write CD-R media\n",
            "  Does read CD-RW media\n",
            "  Does write CD-RW media\n",
            "  Does not read DVD-ROM media\n",
            "  Does not read DVD-R media\n",
            "  Does not write DVD-R media\n",
            "  Does not read DVD-RAM media\n",
            "  Does not write DVD-RAM media\n",
            "  Does support test writing\n",
            "\n",
            "  Does read Mode 2 Form 1 blocks\n",
            "  Does read Mode 2 Form 2 blocks\n",
            "  Does read digital audio blocks\n",
            "  Does restart non-streamed digital audio reads accurately\n",
            "  Does not support BURN-Proof (Sanyo)\n",
            "  Does read multi-session CDs\n",
            "  Does read fixed-packet CD media using Method 2\n",
            "  Does not read CD bar code\n",
            "  Does not read R-W subcode information\n",
            "  Does read raw P-W subcode data from lead in\n",
            "  Does return CD media catalog number\n",
            "  Does return CD ISRC information\n",
            "  Does not support C2 error pointers\n",
            "  Does not deliver composite A/V data\n",
            "\n",
            "  Does play audio CDs\n",
            "  Number of volume control levels: 256\n",
            "  Does support individual volume control setting for each channel\n",
            "  Does support independent mute setting for each channel\n",
            "  Does not support digital output on port 1\n",
            "  Does not support digital output on port 2\n",
            "\n",
            "  Loading mechanism type: tray\n",
            "  Does support ejection of CD via START/STOP command\n",
            "  Does not lock media on power up via prevent jumper\n",
            "  Does allow media to be locked in the drive via PREVENT/ALLOW command\n",
            "  Is not currently in a media-locked state\n",
            "  Does not support changing side of disk\n",
            "  Does not have load-empty-slot-in-changer feature\n",
            "  Does not support Individual Disk Present feature\n",
            "\n",
            "  Maximum read  speed in kB/s: 5645\n",
            "  Current read  speed in kB/s: 3528\n",
            "  Maximum write speed in kB/s: 1411\n",
            "  Current write speed in kB/s: 706\n",
        ]
        (
            deviceType,
            deviceVendor,
            deviceId,
            deviceBufferSize,
            deviceSupportsMulti,
            deviceHasTray,
            deviceCanEject,
        ) = CdWriter._parsePropertiesOutput(output)
        self.assertEqual("Removable CD-ROM", deviceType)
        self.assertEqual("SONY", deviceVendor)
        self.assertEqual("CD-RW  CRX140E", deviceId)
        self.assertEqual(None, deviceBufferSize)
        self.assertEqual(True, deviceSupportsMulti)
        self.assertEqual(True, deviceHasTray)
        self.assertEqual(True, deviceCanEject)

    def testParseOutput_017(self):
        """
        Test _parsePropertiesOutput() for valid data, taken from a real example,
        including stderr and stdout mixed together, "supports multi" removed.
        """
        output = [
            "scsidev: '0,0,0'\n",
            "scsibus: 0 target: 0 lun: 0\n",
            "Linux sg driver version: 3.1.22\n",
            "Cdrecord 1.10 (i686-pc-linux-gnu) Copyright (C) 1995-2001 J\xf6rg Schilling\n",
            "Using libscg version 'schily-0.5'\n",
            "Device type    : Removable CD-ROM\n",
            "Version        : 0\n",
            "Response Format: 1\n",
            "Vendor_info    : 'SONY    '\n",
            "Identifikation : 'CD-RW  CRX140E  '\n",
            "Revision       : '1.0n'\n",
            "Device seems to be: Generic mmc CD-RW.\n",
            "\n",
            "Drive capabilities, per page 2A:\n",
            "\n",
            "  Does read CD-R media\n",
            "  Does write CD-R media\n",
            "  Does read CD-RW media\n",
            "  Does write CD-RW media\n",
            "  Does not read DVD-ROM media\n",
            "  Does not read DVD-R media\n",
            "  Does not write DVD-R media\n",
            "  Does not read DVD-RAM media\n",
            "  Does not write DVD-RAM media\n",
            "  Does support test writing\n",
            "\n",
            "  Does read Mode 2 Form 1 blocks\n",
            "  Does read Mode 2 Form 2 blocks\n",
            "  Does read digital audio blocks\n",
            "  Does restart non-streamed digital audio reads accurately\n",
            "  Does not support BURN-Proof (Sanyo)\n",
            "  Does read fixed-packet CD media using Method 2\n",
            "  Does not read CD bar code\n",
            "  Does not read R-W subcode information\n",
            "  Does read raw P-W subcode data from lead in\n",
            "  Does return CD media catalog number\n",
            "  Does return CD ISRC information\n",
            "  Does not support C2 error pointers\n",
            "  Does not deliver composite A/V data\n",
            "\n",
            "  Does play audio CDs\n",
            "  Number of volume control levels: 256\n",
            "  Does support individual volume control setting for each channel\n",
            "  Does support independent mute setting for each channel\n",
            "  Does not support digital output on port 1\n",
            "  Does not support digital output on port 2\n",
            "\n",
            "  Loading mechanism type: tray\n",
            "  Does support ejection of CD via START/STOP command\n",
            "  Does not lock media on power up via prevent jumper\n",
            "  Does allow media to be locked in the drive via PREVENT/ALLOW command\n",
            "  Is not currently in a media-locked state\n",
            "  Does not support changing side of disk\n",
            "  Does not have load-empty-slot-in-changer feature\n",
            "  Does not support Individual Disk Present feature\n",
            "\n",
            "  Maximum read  speed in kB/s: 5645\n",
            "  Current read  speed in kB/s: 3528\n",
            "  Maximum write speed in kB/s: 1411\n",
            "  Current write speed in kB/s: 706\n",
            "  Buffer size in KB: 4096\n",
        ]
        (
            deviceType,
            deviceVendor,
            deviceId,
            deviceBufferSize,
            deviceSupportsMulti,
            deviceHasTray,
            deviceCanEject,
        ) = CdWriter._parsePropertiesOutput(output)
        self.assertEqual("Removable CD-ROM", deviceType)
        self.assertEqual("SONY", deviceVendor)
        self.assertEqual("CD-RW  CRX140E", deviceId)
        self.assertEqual(4096.0 * 1024.0, deviceBufferSize)
        self.assertEqual(False, deviceSupportsMulti)
        self.assertEqual(True, deviceHasTray)
        self.assertEqual(True, deviceCanEject)

    def testParseOutput_018(self):
        """
        Test _parsePropertiesOutput() for valid data, taken from a real example,
        including stderr and stdout mixed together, "has tray" removed.
        """
        output = [
            "scsidev: '0,0,0'\n",
            "scsibus: 0 target: 0 lun: 0\n",
            "Linux sg driver version: 3.1.22\n",
            "Cdrecord 1.10 (i686-pc-linux-gnu) Copyright (C) 1995-2001 J\xf6rg Schilling\n",
            "Using libscg version 'schily-0.5'\n",
            "Device type    : Removable CD-ROM\n",
            "Version        : 0\n",
            "Response Format: 1\n",
            "Vendor_info    : 'SONY    '\n",
            "Identifikation : 'CD-RW  CRX140E  '\n",
            "Revision       : '1.0n'\n",
            "Device seems to be: Generic mmc CD-RW.\n",
            "\n",
            "Drive capabilities, per page 2A:\n",
            "\n",
            "  Does read CD-R media\n",
            "  Does write CD-R media\n",
            "  Does read CD-RW media\n",
            "  Does write CD-RW media\n",
            "  Does not read DVD-ROM media\n",
            "  Does not read DVD-R media\n",
            "  Does not write DVD-R media\n",
            "  Does not read DVD-RAM media\n",
            "  Does not write DVD-RAM media\n",
            "  Does support test writing\n",
            "\n",
            "  Does read Mode 2 Form 1 blocks\n",
            "  Does read Mode 2 Form 2 blocks\n",
            "  Does read digital audio blocks\n",
            "  Does restart non-streamed digital audio reads accurately\n",
            "  Does not support BURN-Proof (Sanyo)\n",
            "  Does read multi-session CDs\n",
            "  Does read fixed-packet CD media using Method 2\n",
            "  Does not read CD bar code\n",
            "  Does not read R-W subcode information\n",
            "  Does read raw P-W subcode data from lead in\n",
            "  Does return CD media catalog number\n",
            "  Does return CD ISRC information\n",
            "  Does not support C2 error pointers\n",
            "  Does not deliver composite A/V data\n",
            "\n",
            "  Does play audio CDs\n",
            "  Number of volume control levels: 256\n",
            "  Does support individual volume control setting for each channel\n",
            "  Does support independent mute setting for each channel\n",
            "  Does not support digital output on port 1\n",
            "  Does not support digital output on port 2\n",
            "\n",
            "  Does support ejection of CD via START/STOP command\n",
            "  Does not lock media on power up via prevent jumper\n",
            "  Does allow media to be locked in the drive via PREVENT/ALLOW command\n",
            "  Is not currently in a media-locked state\n",
            "  Does not support changing side of disk\n",
            "  Does not have load-empty-slot-in-changer feature\n",
            "  Does not support Individual Disk Present feature\n",
            "\n",
            "  Maximum read  speed in kB/s: 5645\n",
            "  Current read  speed in kB/s: 3528\n",
            "  Maximum write speed in kB/s: 1411\n",
            "  Current write speed in kB/s: 706\n",
            "  Buffer size in KB: 4096\n",
        ]
        (
            deviceType,
            deviceVendor,
            deviceId,
            deviceBufferSize,
            deviceSupportsMulti,
            deviceHasTray,
            deviceCanEject,
        ) = CdWriter._parsePropertiesOutput(output)
        self.assertEqual("Removable CD-ROM", deviceType)
        self.assertEqual("SONY", deviceVendor)
        self.assertEqual("CD-RW  CRX140E", deviceId)
        self.assertEqual(4096.0 * 1024.0, deviceBufferSize)
        self.assertEqual(True, deviceSupportsMulti)
        self.assertEqual(False, deviceHasTray)
        self.assertEqual(True, deviceCanEject)

    def testParseOutput_019(self):
        """
        Test _parsePropertiesOutput() for valid data, taken from a real example,
        including stderr and stdout mixed together, "can eject" removed.
        """
        output = [
            "scsidev: '0,0,0'\n",
            "scsibus: 0 target: 0 lun: 0\n",
            "Linux sg driver version: 3.1.22\n",
            "Cdrecord 1.10 (i686-pc-linux-gnu) Copyright (C) 1995-2001 J\xf6rg Schilling\n",
            "Using libscg version 'schily-0.5'\n",
            "Device type    : Removable CD-ROM\n",
            "Version        : 0\n",
            "Response Format: 1\n",
            "Vendor_info    : 'SONY    '\n",
            "Identifikation : 'CD-RW  CRX140E  '\n",
            "Revision       : '1.0n'\n",
            "Device seems to be: Generic mmc CD-RW.\n",
            "\n",
            "Drive capabilities, per page 2A:\n",
            "\n",
            "  Does read CD-R media\n",
            "  Does write CD-R media\n",
            "  Does read CD-RW media\n",
            "  Does write CD-RW media\n",
            "  Does not read DVD-ROM media\n",
            "  Does not read DVD-R media\n",
            "  Does not write DVD-R media\n",
            "  Does not read DVD-RAM media\n",
            "  Does not write DVD-RAM media\n",
            "  Does support test writing\n",
            "\n",
            "  Does read Mode 2 Form 1 blocks\n",
            "  Does read Mode 2 Form 2 blocks\n",
            "  Does read digital audio blocks\n",
            "  Does restart non-streamed digital audio reads accurately\n",
            "  Does not support BURN-Proof (Sanyo)\n",
            "  Does read multi-session CDs\n",
            "  Does read fixed-packet CD media using Method 2\n",
            "  Does not read CD bar code\n",
            "  Does not read R-W subcode information\n",
            "  Does read raw P-W subcode data from lead in\n",
            "  Does return CD media catalog number\n",
            "  Does return CD ISRC information\n",
            "  Does not support C2 error pointers\n",
            "  Does not deliver composite A/V data\n",
            "\n",
            "  Does play audio CDs\n",
            "  Number of volume control levels: 256\n",
            "  Does support individual volume control setting for each channel\n",
            "  Does support independent mute setting for each channel\n",
            "  Does not support digital output on port 1\n",
            "  Does not support digital output on port 2\n",
            "\n",
            "  Loading mechanism type: tray\n",
            "  Does not lock media on power up via prevent jumper\n",
            "  Does allow media to be locked in the drive via PREVENT/ALLOW command\n",
            "  Is not currently in a media-locked state\n",
            "  Does not support changing side of disk\n",
            "  Does not have load-empty-slot-in-changer feature\n",
            "  Does not support Individual Disk Present feature\n",
            "\n",
            "  Maximum read  speed in kB/s: 5645\n",
            "  Current read  speed in kB/s: 3528\n",
            "  Maximum write speed in kB/s: 1411\n",
            "  Current write speed in kB/s: 706\n",
            "  Buffer size in KB: 4096\n",
        ]
        (
            deviceType,
            deviceVendor,
            deviceId,
            deviceBufferSize,
            deviceSupportsMulti,
            deviceHasTray,
            deviceCanEject,
        ) = CdWriter._parsePropertiesOutput(output)
        self.assertEqual("Removable CD-ROM", deviceType)
        self.assertEqual("SONY", deviceVendor)
        self.assertEqual("CD-RW  CRX140E", deviceId)
        self.assertEqual(4096.0 * 1024.0, deviceBufferSize)
        self.assertEqual(True, deviceSupportsMulti)
        self.assertEqual(True, deviceHasTray)
        self.assertEqual(False, deviceCanEject)

    def testParseOutput_020(self):
        """
        Test _parsePropertiesOutput() for nonsensical data, just a bunch of empty
        lines.
        """
        output = [
            "\n",
            "\n",
            "\n",
            "\n",
            "\n",
            "\n",
            "\n",
            "\n",
            "\n",
            "\n",
            "\n",
            "\n",
            "\n",
            "\n",
            "\n",
        ]
        (
            deviceType,
            deviceVendor,
            deviceId,
            deviceBufferSize,
            deviceSupportsMulti,
            deviceHasTray,
            deviceCanEject,
        ) = CdWriter._parsePropertiesOutput(output)
        self.assertEqual(None, deviceType)
        self.assertEqual(None, deviceVendor)
        self.assertEqual(None, deviceId)
        self.assertEqual(None, deviceBufferSize)
        self.assertEqual(False, deviceSupportsMulti)
        self.assertEqual(False, deviceHasTray)
        self.assertEqual(False, deviceCanEject)

    def testParseOutput_021(self):
        """
        Test _parsePropertiesOutput() for nonsensical data, just an empty list.
        """
        output = []
        (
            deviceType,
            deviceVendor,
            deviceId,
            deviceBufferSize,
            deviceSupportsMulti,
            deviceHasTray,
            deviceCanEject,
        ) = CdWriter._parsePropertiesOutput(output)
        self.assertEqual(None, deviceType)
        self.assertEqual(None, deviceVendor)
        self.assertEqual(None, deviceId)
        self.assertEqual(None, deviceBufferSize)
        self.assertEqual(False, deviceSupportsMulti)
        self.assertEqual(False, deviceHasTray)
        self.assertEqual(False, deviceCanEject)



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
#              C E D A R
#          S O L U T I O N S       "Software done right."
#           S O F T W A R E
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Copyright (c) 2004-2006,2008,2010,2015 Kenneth J. Pronovici.
# All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License,
# Version 2, as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# Copies of the GNU General Public License are available from
# the Free Software Foundation website, http://www.gnu.org/.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Author   : Kenneth J. Pronovici <pronovic@ieee.org>
# Language : Python 3
# Project  : Cedar Backup, release 3
# Purpose  : Provides unit-testing utilities.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

########################################################################
# Module documentation
########################################################################

"""
Provides unit-testing utilities.

These utilities are kept here, separate from util.py, because they provide
common functionality that I do not want exported "publicly" once Cedar Backup
is installed on a system.  They are only used for unit testing, and are only
useful within the source tree.

Many of these functions are in here because they are "good enough" for unit
test work but are not robust enough to be real public functions.  Others (like
:any:`removedir`) do what they are supposed to, but I don't want responsibility for
making them available to others.

:author: Kenneth J. Pronovici <pronovic@ieee.org>
"""


########################################################################
# Imported modules
########################################################################

import getpass
import logging
import os
import platform
import random
import string  # pylint: disable=W0402
import sys
import tarfile
import time
from io import StringIO

from CedarBackup3.cli import setupPathResolver
from CedarBackup3.config import Config, OptionsConfig
from CedarBackup3.customize import customizeOverrides
from CedarBackup3.util import encodePath, executeCommand, nullDevice, pathJoin

########################################################################
# Public functions
########################################################################

##############################
# setupDebugLogger() function
##############################


def setupDebugLogger():
    """
    Sets up a screen logger for debugging purposes.

    Normally, the CLI functionality configures the logger so that
    things get written to the right place.  However, for debugging
    it's sometimes nice to just get everything -- debug information
    and output -- dumped to the screen.  This function takes care
    of that.
    """
    logger = logging.getLogger("CedarBackup3")
    logger.setLevel(logging.DEBUG)  # let the logger see all messages
    formatter = logging.Formatter(fmt="%(message)s")
    handler = logging.StreamHandler(stream=sys.stdout)
    handler.setFormatter(formatter)
    handler.setLevel(logging.DEBUG)
    logger.addHandler(handler)


def configureLogging():
    """Optionally disable system logging based on configuration in the environment."""
    if "FULL_LOGGING" not in os.environ or os.environ["FULL_LOGGING"] == "N":
        devnull = nullDevice()
        handler = logging.FileHandler(filename=devnull)
        handler.setLevel(logging.NOTSET)
        logger = logging.getLogger("CedarBackup3")
        logger.setLevel(logging.NOTSET)
        logger.addHandler(handler)


#################
# setupOverrides
#################


def setupOverrides():
    """
    Set up any platform-specific overrides that might be required.

    When packages are built, this is done manually (hardcoded) in customize.py
    and the overrides are set up in cli.cli().  This way, no runtime checks need
    to be done.  This is safe, because the package maintainer knows exactly
    which platform (Debian or not) the package is being built for.

    Unit tests are different, because they might be run anywhere.  So, we
    attempt to make a guess about plaform using platformDebian(), and use that
    to set up the custom overrides so that platform-specific unit tests continue
    to work.
    """
    config = Config()
    config.options = OptionsConfig()
    if platformDebian():
        customizeOverrides(config, platform="debian")
    else:
        customizeOverrides(config, platform="standard")
    setupPathResolver(config)


###########################
# findResources() function
###########################


def findResources(resources, dataDirs):
    """
    Returns a dictionary of locations for various resources.
    Args:
       resources: List of required resources
       dataDirs: List of data directories to search within for resources
    Returns:
        Dictionary mapping resource name to resource path
    Raises:
       Exception: If some resource cannot be found
    """
    mapping = {}
    for resource in resources:
        for resourceDir in dataDirs:
            path = pathJoin(resourceDir, resource)
            if os.path.exists(path):
                mapping[resource] = path
                break
        else:
            raise Exception("Unable to find resource [%s]." % resource)
    return mapping


##############################
# commandAvailable() function
##############################


def commandAvailable(command):
    """
    Indicates whether a command is available on $PATH somewhere.
    This should work on both Windows and UNIX platforms.
    Args:
       command: Commang to search for
    Returns:
        Boolean true/false depending on whether command is available
    """
    if "PATH" in os.environ:
        for path in os.environ["PATH"].split(os.sep):
            if os.path.exists(pathJoin(path, command)):
                return True
    return False


#######################
# buildPath() function
#######################


def buildPath(components):
    """
    Builds a complete path from a list of components.
    For instance, constructs ``"/a/b/c"`` from ``["/a", "b", "c"]``.
    Args:
       components: List of components
    Returns:
        String path constructed from components
    Raises:
       ValueError: If a path cannot be encoded properly
    """
    path = components[0]
    for component in components[1:]:
        path = pathJoin(path, component)
    return encodePath(path)


#######################
# removedir() function
#######################


def removedir(tree):
    """
    Recursively removes an entire directory.
    This is basically taken from an example on python.com.
    Args:
       tree: Directory tree to remove
    Raises:
       ValueError: If a path cannot be encoded properly
    """
    tree = encodePath(tree)
    for root, dirs, files in os.walk(tree, topdown=False):
        for name in files:
            path = pathJoin(root, name)
            if os.path.islink(path):
                os.remove(path)
            elif os.path.isfile(path):
                os.remove(path)
        for name in dirs:
            path = pathJoin(root, name)
            if os.path.islink(path):
                os.remove(path)
            elif os.path.isdir(path):
                os.rmdir(path)
    os.rmdir(tree)


########################
# extractTar() function
########################


def extractTar(tmpdir, filepath):
    """
    Extracts the indicated tar file to the indicated tmpdir.
    Args:
       tmpdir: Temp directory to extract to
       filepath: Path to tarfile to extract
    Raises:
       ValueError: If a path cannot be encoded properly
    """
    # pylint: disable=E1101
    tmpdir = encodePath(tmpdir)
    filepath = encodePath(filepath)
    with tarfile.open(filepath) as tar:
        try:
            tar.format = tarfile.GNU_FORMAT
        except AttributeError:
            tar.posix = False
        for tarinfo in tar:
            # Starting in Python 3.12, we have to explicitly specify filter behavior
            # See: https://docs.python.org/3/library/tarfile.html#supporting-older-python-versions
            if hasattr(tarfile, "data_filter"):
                # noinspection PyArgumentList
                tar.extract(member=tarinfo, path=tmpdir, filter="data")
            else:
                tar.extract(member=tarinfo, path=tmpdir)


###########################
# changeFileAge() function
###########################


def changeFileAge(filename, subtract=None):
    """
    Changes a file age using the ``os.utime`` function.

    *Note:* Some platforms don't seem to be able to set an age precisely.  As a
    result, whereas we might have intended to set an age of 86400 seconds, we
    actually get an age of 86399.375 seconds.  When util.calculateFileAge()
    looks at that the file, it calculates an age of 0.999992766204 days, which
    then gets truncated down to zero whole days.  The tests get very confused.
    To work around this, I always subtract off one additional second as a fudge
    factor.  That way, the file age will be *at least* as old as requested
    later on.

    Args:
       filename: File to operate on
       subtract: Number of seconds to subtract from the current time
    Raises:
       ValueError: If a path cannot be encoded properly
    """
    filename = encodePath(filename)
    newTime = time.time() - 1
    if subtract is not None:
        newTime -= subtract
    os.utime(filename, (newTime, newTime))


###########################
# getMaskAsMode() function
###########################


def getMaskAsMode():
    """
    Returns the user's current umask inverted to a mode.
    A mode is mostly a bitwise inversion of a mask, i.e. mask 002 is mode 775.
    Returns:
        Umask converted to a mode, as an integer
    """
    umask = os.umask(0o777)
    os.umask(umask)
    return int(~umask & 0o777)  # invert, then use only lower bytes


######################
# getLogin() function
######################


def getLogin():
    """
    Returns the name of the currently-logged in user.  This might fail under
    some circumstances - but if it does, our tests would fail anyway.
    """
    return getpass.getuser()


############################
# randomFilename() function
############################


# noinspection PyTypeChecker
def randomFilename(length, prefix=None, suffix=None):
    """
    Generates a random filename with the given length.
    @return Random filename
    """
    characters = [None] * length
    for i in range(length):
        characters[i] = random.choice(string.ascii_uppercase)
    if prefix is None:
        prefix = ""
    if suffix is None:
        suffix = ""
    return "%s%s%s" % (prefix, "".join(characters), suffix)


####################################
# failUnlessAssignRaises() function
####################################


# pylint: disable=W0613
def failUnlessAssignRaises(testCase, exception, obj, prop, value):
    """
    Equivalent of ``failUnlessRaises``, but used for property assignments instead.

    It's nice to be able to use ``failUnlessRaises`` to check that a method call
    raises the exception that you expect.  Unfortunately, this method can't be
    used to check Python propery assignments, even though these property
    assignments are actually implemented underneath as methods.

    This function (which can be easily called by unit test classes) provides an
    easy way to wrap the assignment checks.  It's not pretty, or as intuitive as
    the original check it's modeled on, but it does work.

    Let's assume you make this method call::

       testCase.failUnlessAssignRaises(ValueError, collectDir, "absolutePath", absolutePath)

    If you do this, a test case failure will be raised unless the assignment::

       collectDir.absolutePath = absolutePath

    fails with a ``ValueError`` exception.  The failure message differentiates
    between the case where no exception was raised and the case where the wrong
    exception was raised.

    *Note:* Internally, the ``missed`` and ``instead`` variables are used rather
    than directly calling ``testCase.fail`` upon noticing a problem because the
    act of "failure" itself generates an exception that would be caught by the
    general ``except`` clause.

    Args:
       testCase: PyUnit test case object (i.e. self)
       exception: Exception that is expected to be raised
       obj: Object whose property is to be assigned to
       prop: Name of the property, as a string
       value: Value that is to be assigned to the property

    @see: ``unittest.TestCase.failUnlessRaises``
    """
    missed = False
    instead = None
    try:
        exec("obj.%s = value" % prop)  # pylint: disable=W0122
        missed = True
    except exception:
        pass
    except Exception as e:
        instead = e
    if missed:
        testCase.fail("Expected assignment to raise %s, but got no exception." % (exception.__name__))
    if instead is not None:
        testCase.fail("Expected assignment to raise %s, but got %s instead." % (ValueError, instead.__class__.__name__))


###########################
# captureOutput() function
###########################


def captureOutput(c):
    """
    Captures the output (stdout, stderr) of a function or a method.

    Some of our functions don't do anything other than just print output.  We
    need a way to test these functions (at least nominally) but we don't want
    any of the output spoiling the test suite output.

    This function just creates a dummy file descriptor that can be used as a
    target by the callable function, rather than ``stdout`` or ``stderr``.

    *Note:* This method assumes that ``callable`` doesn't take any arguments
    besides keyword argument ``fd`` to specify the file descriptor.

    Args:
       c: Callable function or method

    Returns:
        Output of function, as one big string
    """
    fd = StringIO()
    c(fd=fd)
    result = fd.getvalue()
    fd.close()
    return result


#########################
# _isPlatform() function
#########################


def _isPlatform(name):
    """
    Returns boolean indicating whether we're running on the indicated platform.
    Args:
       name: Platform name to check, currently one of "windows" or "macosx"
    """
    if name == "windows":
        return sys.platform == "win32"
    elif name == "macosx":
        return sys.platform == "darwin"
    elif name == "debian":
        return platform.platform(False, False).find("debian") > 0
    elif name == "cygwin":
        return platform.platform(True, True).startswith("CYGWIN")
    else:
        raise ValueError("Unknown platform [%s]." % name)


############################
# platformDebian() function
############################


def platformDebian():
    """
    Returns boolean indicating whether this is the Debian platform.
    """
    return _isPlatform("debian")


############################
# platformMacOsX() function
############################


def platformMacOsX():
    """
    Returns boolean indicating whether this is the Mac OS X platform.
    """
    return _isPlatform("macosx")


#############################
# platformWindows() function
#############################


def platformWindows():
    """
    Returns boolean indicating whether this is the Windows platform.
    """
    return _isPlatform("windows")


####################################
# platformSupportsLinks() function
####################################


def platformSupportsLinks():
    """Whether the platform supports soft links"""
    return not platformWindows()


##############################
# availableLocales() function
##############################


def availableLocales():
    """
    Returns a list of available locales on the system
    Returns:
        List of string locale names
    """
    locales = []
    output = executeCommand(["locale"], ["-a"], returnOutput=True, ignoreStderr=True)[1]
    for line in output:
        locales.append(line.rstrip())  # pylint: disable=E1101
    return locales
